import logging
logger = logging.getLogger("panwid.tabview")

import urwid
from urwid_utils.palette import *

# Based on the tabview widget from github/@ountainstorm
# https://github.com/mountainstorm/mt_urwid/

class TabHandle(urwid.WidgetWrap):

    LABEL_CHARS_UNLOCKED = u"\u24e7 "
    LABEL_CHARS_LOCKED = u"\u25cb "

    def __init__(self, tab_view, title, locked=False, padding = 3,
                 attr_inactive = {}, attr_active = {}):
        self.tab_view = tab_view
        self.title = title
        self.locked = locked
        if locked:
            self.label = self.LABEL_CHARS_LOCKED + title
        else:
            self.label = self.LABEL_CHARS_UNLOCKED + title

        self.label = ' '*padding + self.label + ' '*padding

        self.text = urwid.SelectableIcon(self.label)
        # self.padding = urwid.Padding(self.text, align="left", width=20, left=3, right=3)
        self.attr = urwid.AttrMap(self.text, attr_inactive, "tabview_active")
        super(TabHandle, self).__init__(self.attr)

    def set_text_attr(self, attr):
        self.text.set_text((attr, self.label))


    def selectable(self):
        return True

    def keypress(self, size, key):

        if key == "enter":
            self.tab_view._set_active_by_tab(self)
        elif key == 'tab':
            self.tab_view.set_active_next()
        elif key == 'shift tab':
            self.tab_view.set_active_prev()
        else:
            return key

    def mouse_event(self, size, event, button, col, row, focus):
        if button == 1:
            if not self.locked:
                tab_width = self._w.pack(size)[0]
                if col <= 2:
                    # close this tab
                    self.tab_view._close_by_tab(self)
                    return
            # make this tab active
            self.tab_view._set_active_by_tab(self)
            #raise AttributeError("b: %s - c: %u, r: %u - ev: %s" % (button, col, row, event))


class TabHeader(urwid.WidgetWrap):

    def __init__(self, attr_inactive={}, attr_active={}, divider = True):

        self.columns = urwid.Columns([], 1)

        contents = [ ('weight', 1, x) for x in [self.columns] ]
        # contents = [ ('pack', self.attr)  ]
        if divider:
            contents += [ ('pack', urwid.Divider('-')) ]
        self.pile = urwid.Pile(contents)
        # self.pile.selectable = lambda: True
        super(TabHeader, self).__init__(self.pile)

    @property
    def contents(self):
        return self.columns.contents

    def set_focus(self, idx):

        self.columns.focus_position = idx

    def options(self, s):
        return self.columns.options(s)


class Tab(object):

    HOTKEYS = dict()

    def __init__(self, label, content, hotkey=None, locked=False):

        self.label = label
        self.content = content
        if hotkey:
            Tab.HOTKEYS["meta %s" %(hotkey)] = self
        else:
            hotkey = "meta %s" %(self.label[0].lower())
            if not hotkey in Tab.HOTKEYS:
                Tab.HOTKEYS[hotkey] = self
            else:
                hotkey = None
        self.hotkey = hotkey
        self.locked = locked

    # yuck
    def __getitem__(self, idx):
        if idx == "0":
            return self.label
        elif idx == 1:
            return self.content



class TabView(urwid.WidgetWrap):

    signals = ["activate"]

    def __init__(self, tabs,
                 attr_inactive={None: "tabview_inactive"},
                 attr_active={None: "tabview_active"},
                 selected = 0, tab_bar_initial_focus = False):
        self.attr_inactive = attr_inactive
        self.attr_active = attr_active
        self._contents = []
        self.tab_bar = TabHeader(attr_inactive, attr_active)
        self.body = urwid.AttrMap(urwid.SolidFill(' '), attr_active)
        display_widget = urwid.Pile(
            ( ('pack', self.tab_bar), ('weight', 1, self.body) )
        )
        # display_widget.selectable = lambda: True
        super(TabView, self).__init__(display_widget)

        if not tab_bar_initial_focus:
            display_widget.focus_position = 1

        # now add all the tabs
        for tab in tabs:
            self.add_tab(tab)
        if selected is not None:
            self.set_active_tab(selected)

    @property
    def active_tab(self):
        return self._contents[self.active_tab_idx]

    @classmethod
    def get_palette_entries(cls):
        return {
            "tabview_inactive": PaletteEntry(
                foreground = "light gray",
                background = "black"
            ),
            "tabview_active": PaletteEntry(
                foreground = "white",
                background = "dark blue",
                foreground_high = "white",
                background_high = "#009",
            ),
        }

    def add_tab(self, tab):
        label = tab.label
        content = tab.content
        hokey = tab.hotkey
        locked = tab.locked

        self.tab_bar.contents.append(
            (
                TabHandle(
                    self,
                    label,
                    locked,
                    attr_active = self.attr_active,
                    attr_inactive = self.attr_inactive,

                ),
                self.tab_bar.options('pack')
            )
        )
        self._contents.append(tab)
        self.set_active_tab(len(self._contents)-1)

    def set_active_tab(self, idx):

        if idx < 0 or idx > len(self._contents) - 1:
            return

        self.tab_bar.set_focus(idx)

        for i, tab in enumerate(self.tab_bar.contents):
            if i == idx:
                tab[0].set_text_attr(self.attr_active[None])
            else:
                tab[0].set_text_attr(self.attr_inactive[None])

        self._w.contents[1] = (
            urwid.AttrMap(
                self._contents[idx].content,
                self.attr_inactive
            ),
            self._w.contents[1][1]
        )
        self.active_tab_idx = idx
        urwid.signals.emit_signal(self, "activate", self, self._contents[idx])

    def get_tab_by_label(self, label):

        for tab in self._contents:
            if tab.label == label:
                return tab.content
        return None

    def get_tab_index_by_label(self, label):

        for i, tab in enumerate(self._contents):
            if tab.label == label:
                return i
        return None


    def set_active_next(self):
        if self.active_tab_idx < (len(self._contents)-1):
            self.set_active_tab(self.active_tab_idx+1)
        else:
            self.set_active_tab(0)

    def set_active_prev(self):
        if self.active_tab_idx > 0:
            self.set_active_tab(self.active_tab_idx-1)
        else:
            self.set_active_tab(len(self._contents)-1)

    def close_active_tab(self):
        if not self.tab_bar.contents[self.active_tab_idx][0].locked:
            del self.tab_bar.contents[self.active_tab_idx]
            new_idx = self.active_tab_idx
            if len(self._contents) <= self.active_tab_idx:
                new_idx -= 1
            del self._contents[self.active_tab_idx]
            self.set_active_tab(new_idx)

    def _set_active_by_tab(self, tab):
        for idx, t in enumerate(self.tab_bar.contents):
            if t[0] is tab:
                self.set_active_tab(idx)
                break

    def _close_by_tab(self, tab):
        for idx, t in enumerate(self.tab_bar.contents):
            if t[0] is tab:
                self.set_active_tab(idx)
                self.close_active_tab()
                break

    def keypress(self, size, key):

        if key in Tab.HOTKEYS:
            idx = self.get_tab_index_by_label(Tab.HOTKEYS[key].label)
            self.set_active_tab(idx)
        else:
            return super(TabView, self).keypress(size, key)

__all__ = ["TabView", "Tab"]



import logging
logger = logging.getLogger(__name__)

class HighlightableTextMixin(object):

    @property
    def highlight_state(self):
        if not getattr(self, "_highlight_state", False):
            self._highlight_state = False
            self._highlight_case_sensitive = False
            self._highlight_string = None
        return self._highlight_state

    @property
    def highlight_content(self):
        if self.highlight_state:
            return self.get_highlight_text()
        else:
            return self.highlight_source


    def highlight(self, start, end):
        self._highlight_state = True
        self._highlight_location = (start, end)
        self.on_highlight()

    def unhighlight(self):
        self._highlight_state = False
        self._highlight_location = None
        self.on_unhighlight()

    def get_highlight_text(self):

        if not self._highlight_location:
            return None

        return [
            (self.highlightable_attr_normal, self.highlight_source[:self._highlight_location[0]]),
            (self.highlightable_attr_highlight, self.highlight_source[self._highlight_location[0]:self._highlight_location[1]]),
            (self.highlightable_attr_normal, self.highlight_source[self._highlight_location[1]:]),
        ]

    @property
    def highlight_source(self):
        raise NotImplementedError

    @property
    def highlightable_attr_normal(self):
        raise NotImplementedError

    @property
    def highlightable_attr_highlight(self):
        raise NotImplementedError

    def on_highlight(self):
        pass

    def on_unhighlight(self):
        pass

__all__ = ["HighlightableTextMixin"]



import logging
logger = logging.getLogger("panwid.datatable")
import raccoon as rc
import collections

class DataTableDataFrame(rc.DataFrame):

    DATA_TABLE_COLUMNS = ["_dirty", "_focus_position", "_value_fn", "_cls", "_details", "_rendered_row"]

    def __init__(self, data=None, columns=None, index=None, index_name="index", sort=None):

        self.sidecar_columns = []
        if columns and not index_name in columns:
            columns.insert(0, index_name)
        columns += self.DATA_TABLE_COLUMNS
        super(DataTableDataFrame, self).__init__(
            data=data,
            columns=columns,
            index=index,
            index_name=index_name,
            sort=sort
        )
        # for c in self.DATA_TABLE_COLUMNS:
        #     self[c] = None

    def _validate_index(self, indexes):
        try:
            return super(DataTableDataFrame, self)._validate_index(indexes)
        except ValueError:
            logger.error("duplicates in index: %s" %(
                [item for item, count
                 in list(collections.Counter(indexes).items()) if count > 1
                ]))
            raise


    def log_dump(self, n=5, columns=None, label=None):
        df = self
        if columns:
            if not isinstance(columns, list):
                columns = [columns]
            df = df[columns]
        logger.info("%slength: %d, index: %s [%s%s]\n%s" %(
            "%s, " %(label) if label else "",
            len(self),
            self.index_name,
            ",".join([str(x) for x in self.index[0:min(n, len(self.index))]]),
            "..." if len(self.index) > n else "",
            df.head(n)))

    @staticmethod
    def extract_keys(obj):
        return obj.keys() if hasattr(obj, "keys") else obj.__dict__.keys()

    @staticmethod
    def extract_value(obj, key):
        if isinstance(obj, collections.abc.MutableMapping):
            # raise Exception(obj)
            return obj.get(key, None)
        else:
            return getattr(obj, key, None)

    def transpose_data(self, rows, with_sidecar = False):

        # raise Exception([ r[self.index_name] for r, s in rows])

        if with_sidecar:
            data_columns, self.sidecar_columns = [
                list(set().union(*x))
                for x in zip(*[(
                        self.extract_keys(d),
                        self.extract_keys(s)
                )
                               for d, s in rows ])
            ]

        else:
            data_columns = list(
                set().union(*(list(d.keys()
                               if hasattr(d, "keys")
                               else d.__dict__.keys())
                          for d in rows))
            )

        data_columns += [
            c for c in self.columns
            if c not in data_columns
            and c not in self.sidecar_columns
            and c != self.index_name
            and c not in self.DATA_TABLE_COLUMNS
        ]
        data_columns += ["_cls"]

        data = dict(
                    list(zip((data_columns + self.sidecar_columns),
                        [ list(z) for z in zip(*[[
                            self.extract_value(d, k) if k in data_columns else self.extract_value(s, k)
                            for k in data_columns + self.sidecar_columns]
                            for d, s in (
                                rows
                                if with_sidecar
                                else [ (r, {})  for r in rows]
                            )
                        ])]
                    ))
                )

        return data


    def update_rows(self, rows, replace=False, with_sidecar = False):

        if not len(rows):
            return []

        data = self.transpose_data(rows, with_sidecar = with_sidecar)
        # data["_details"] = [{"open": False, "disabled": False}] * len(rows)
        data["_cls"] = [type(rows[0][0] if with_sidecar else rows[0])] * len(rows) # all rows assumed to have same class

        # raise Exception(data["_cls"])
        # if not "_details" in data:
        #     data["_details"] = [{"open": False, "disabled": False}] * len(rows)

        if replace:
            if len(rows):
                indexes = [x for x in self.index if x not in data.get(self.index_name, [])]
                if len(indexes):
                    self.delete_rows(indexes)
            else:
                self.delete_all_rows()

            # logger.info(f"update_rowGs: {self.index}, {data[self.index_name]}")

        if self.index_name not in data:
            index = list(range(len(self), len(self) + len(rows)))
            data[self.index_name] = index
        else:
            index = data[self.index_name]

        for c in data.keys():
            # try:
                # raise Exception(data[self.index_name], c, data[c])
                self.set(data[self.index_name], c, data[c])
            # except ValueError as e:
            #     logger.error(e)
            #     logger.info(f"update_rows: {self.index}, {data}")
            #
        for idx in data[self.index_name]:
            if not self.get(idx, "_details"):
                self.set(idx, "_details", {"open": False, "disabled": False})

        return data.get(self.index_name, [])

    def append_rows(self, rows):

        length = len(rows)
        if not length:
            return

        colnames = list(self.columns) + [c for c in self.DATA_TABLE_COLUMNS if c not in self.columns]

        # data_columns = list(set().union(*(list(d.keys()) for d in rows)))
        data = self.transpose_data(rows)
        colnames += [c for c in data.keys() if c not in colnames]

        for c in self.columns:
            if not c in data:
                data[c] = [None]*length

        for c in colnames:
            if not c in self.columns:
                self[c] = None

        kwargs = dict(
            columns =  colnames,
            data = data,
            sort=False,
            index=data[self.index_name],
            index_name = self.index_name,
        )

        try:
            newdata = DataTableDataFrame(**kwargs)
        except ValueError:
            raise Exception(kwargs)
        # newdata.log_dump()
        # self.log_dump(10, label="before")
        try:
            self.append(newdata)
        except ValueError:
            raise Exception(f"{self.index}, {newdata}")
        # self.log_dump(10, label="after")

    # def add_column(self, column, data=None):
    #     self[column] = data

    def clear(self):
        self.delete_all_rows()
        # self.delete_rows(self.index)



from __future__ import division
import logging
logger = logging.getLogger(__name__.split(".")[0])

import urwid
from urwid_utils.palette import *
from .scroll import ScrollBar

class ListBoxScrollBar(urwid.WidgetWrap):

    def __init__(self, parent):
        self.parent = parent
        self.pile = urwid.Pile([])
        super(ListBoxScrollBar, self).__init__(self.pile)

    def update(self, size):
        width, height = size
        scroll_marker_height = 1
        del self.pile.contents[:]
        if (len(self.parent.body)
            and self.parent.row_count
            and self.parent.focus is not None
            and self.parent.row_count > height):
            scroll_position = int(
                self.parent.focus_position / self.parent.row_count * height
            )
            scroll_marker_height = max( height * (height / self.parent.row_count ), 1)
        else:
            scroll_position = 0

        pos_marker = urwid.AttrMap(urwid.Text(" "),
                                   {None: "scroll_pos"}
        )

        down_marker = urwid.AttrMap(urwid.Text(u"\N{DOWNWARDS ARROW}"),
                                   {None: "scroll_marker"}
        )

        begin_marker = urwid.AttrMap(urwid.Text(u"\N{CIRCLED MINUS}"),
                                   {None: "scroll_marker"}
        )

        end_marker = urwid.AttrMap(urwid.Text(u"\N{CIRCLED PLUS}"),
                                   {None: "scroll_marker"}
        )

        view_marker = urwid.AttrMap(urwid.Text(" "),
                                    {None: "scroll_view"}
        )

        bg_marker = urwid.AttrMap(urwid.Text(" "),
                                   {None: "scroll_bg"}
        )

        for i in range(height):
            if abs( i - scroll_position ) <= scroll_marker_height//2:
                if i == 0 and self.parent.focus_position == 0:
                    marker = begin_marker
                elif i+1 == height and self.parent.row_count == self.parent.focus_position+1:
                    marker = end_marker
                elif self.parent.focus_position is not None and len(self.parent.body) == self.parent.focus_position+1 and i == scroll_position + scroll_marker_height//2:
                    marker = down_marker
                else:
                    marker = pos_marker
            else:
                if i < scroll_position:
                    marker = view_marker
                elif self.parent.row_count and i/height < ( len(self.parent.body) / self.parent.row_count):
                    marker = view_marker
                else:
                    marker = bg_marker
            self.pile.contents.append(
                (urwid.Filler(marker), self.pile.options("weight", 1))
            )
        self._invalidate()

    def selectable(self):
        # FIXME: mouse click/drag
        return False

class ScrollingListBox(urwid.WidgetWrap):

    signals = ["select",
               "drag_start", "drag_continue", "drag_stop",
               "load_more"]

    scrollbar_class = ScrollBar

    def __init__(self, body,
                 infinite = False,
                 with_scrollbar=False,
                 row_count_fn = None,
                 thumb_char=None,
                 trough_char=None,
                 thumb_indicator_top=None,
                 thumb_indicator_bottom=None):

        self.infinite = infinite
        self.with_scrollbar = with_scrollbar
        self.row_count_fn = row_count_fn

        self._width = None
        self._height = 0
        self._rows_max = None

        self.mouse_state = 0
        self.drag_from = None
        self.drag_last = None
        self.drag_to = None
        self.load_more = False
        self.page = 0

        self.queued_keypress = None
        w = self.listbox = urwid.ListBox(body)

        self.columns = urwid.Columns([
            ('weight', 1, self.listbox)
        ])
        if self.with_scrollbar:
            self.scroll_bar = ListBoxScrollBar(self)
            self.columns.contents.append(
                (self.scroll_bar, self.columns.options("given", 1))
            )
        super(ScrollingListBox, self).__init__(self.columns)
        urwid.connect_signal(self.body, "modified", self.on_modified)

    def on_modified(self):
        if self.with_scrollbar and len(self.body):
            self.scroll_bar.update(self.size)

    def rows_max(self, size, focus=False):
        return urwid.ListBox.rows_max(self, size, focus)


    @classmethod
    def get_palette_entries(cls):

        return {

            "scroll_pos": PaletteEntry(
                mono = "white",
                foreground = "black",
                background = "white",
                foreground_high = "black",
                background_high = "white"
            ),
            "scroll_marker": PaletteEntry(
                mono = "white,bold",
                foreground = "black,bold",
                background = "white",
                foreground_high = "black,bold",
                background_high = "white"
            ),
            "scroll_view": PaletteEntry(
                mono = "black",
                foreground = "black",
                background = "light gray",
                foreground_high = "black",
                background_high = "g50"
            ),
            "scroll_bg": PaletteEntry(
                mono = "black",
                foreground = "light gray",
                background = "dark gray",
                foreground_high = "light gray",
                background_high = "g23"
            ),

        }

    def mouse_event(self, size, event, button, col, row, focus):

        SCROLL_WHEEL_HEIGHT_RATIO = 0.5
        if row < 0 or row >= self._height or not len(self.listbox.body):
            return
        if event == 'mouse press':
            if button == 1:
                self.mouse_state = 1
                self.drag_from = self.drag_last = (col, row)
            elif button == 4:
                pos = self.listbox.focus_position - int(self._height * SCROLL_WHEEL_HEIGHT_RATIO)
                if pos < 0:
                    pos = 0
                self.listbox.focus_position = pos
                self.listbox.make_cursor_visible(size)
                self._invalidate()
            elif button == 5:
                pos = self.listbox.focus_position + int(self._height * SCROLL_WHEEL_HEIGHT_RATIO)
                if pos > len(self.listbox.body) - 1:
                    if self.infinite:
                        self.load_more = True
                    pos = len(self.listbox.body) - 1
                self.listbox.focus_position = pos
                self.listbox.make_cursor_visible(size)
                self._invalidate()
        elif event == 'mouse drag':
            if self.drag_from is None:
                return
            if button == 1:
                self.drag_to = (col, row)
                if self.mouse_state == 1:
                    self.mouse_state = 2
                    urwid.signals.emit_signal(
                        self, "drag_start",self, self.drag_from
                    )
                else:
                    urwid.signals.emit_signal(
                        self, "drag_continue",self,
                        self.drag_last, self.drag_to
                    )

            self.drag_last = (col, row)

        elif event == 'mouse release':
            if self.mouse_state == 2:
                self.drag_to = (col, row)
                urwid.signals.emit_signal(
                    self, "drag_stop",self, self.drag_from, self.drag_to
                )
            self.mouse_state = 0
        return super(ScrollingListBox, self).mouse_event(size, event, button, col, row, focus)


    def keypress(self, size, key):

        command = self._command_map[key]
        if not command:
            return super(ScrollingListBox, self).keypress(size, key)

        # down, page down at end trigger load of more data
        if (
                command in ["cursor down", "cursor page down"]
                and self.infinite
                and (
                    not len(self.body)
                    or self.focus_position == len(self.body)-1)
        ):
                self.load_more = True
                self.queued_keypress = key
                self._invalidate()

        elif command == "activate":
            urwid.signals.emit_signal(self, "select", self, self.selection)

        # else:
        return super(ScrollingListBox, self).keypress(size, key)

    @property
    def selection(self):

        if len(self.body):
            return self.body[self.focus_position]

    @property
    def size(self):
        return (self._width, self._height)

    def render(self, size, focus=False):

        maxcol = size[0]
        self._width = maxcol
        if len(size) > 1:
            maxrow = size[1]
            modified = self._height == 0
            self._height = maxrow
            if modified:
                self.on_modified()
        else:
            self._height = 0

        # print
        # print
        # print self.listbox.get_focus_offset_inset(size)
        if (self.load_more
            and (len(self.body) == 0
                 or "bottom" in self.ends_visible((maxcol, maxrow))
            )
        ):

            self.load_more = False
            self.page += 1
            # old_len = len(self.body)
            try:
                focus = self.focus_position
            except IndexError:
                focus = None
            urwid.signals.emit_signal(
                self, "load_more", focus)
            if (self.queued_keypress
                and focus is not None
                # and focus < len(self.body)-1
            ):
                # logger.info(f"send queued keypress: {focus}, {len(self.body)}")
                self.keypress(size, self.queued_keypress)
            self.queued_keypress = None
            # self.listbox._invalidate()
            # self._invalidate()

        return super(ScrollingListBox, self).render(size, focus)


    def disable(self):
        self.selectable = lambda: False

    def enable(self):
        self.selectable = lambda: True

    @property
    def contents(self):
        return self.columns.contents

    @property
    def focus(self):
        return self.listbox.focus

    @property
    def focus_position(self):
        if not len(self.listbox.body):
            raise IndexError
        try:
            return self.listbox.focus_position
        except IndexError:
            pass
        return None

    @focus_position.setter
    def focus_position(self, value):
        if not len(self.body):
            return
        self.listbox.focus_position = value
        self.listbox._invalidate()

    def __getattr__(self, attr):
        if attr in ["ends_visible", "focus_position", "set_focus", "set_focus_valign", "body", "focus"]:
            return getattr(self.listbox, attr)
        # elif attr == "body":
        #     return self.walker
        raise AttributeError(attr)

    @property
    def row_count(self):
        if self.row_count_fn:
            return self.row_count_fn()
        return len(self.body)

__all__ = ["ScrollingListBox"]



import unittest

from panwid.datatable import *
from orderedattrdict import AttrDict

class TestDataTableWithIndex(unittest.TestCase):

    def setUp(self):

        self.data = [
            dict(a=1, b=2.345, c="foo"),
            dict(a=2, b=4.817, c="bar"),
            dict(a=3, b=-3.19, c="baz")
        ]
        self.columns = [
            DataTableColumn("a"),
            DataTableColumn("b"),
            DataTableColumn("c")
        ]

    def test_create(self):

        dt = DataTable(self.columns, data=self.data, index="a")
        dt.refresh()
        self.assertEqual(len(dt), 3)

    def test_create_without_index(self):

        dt = DataTable(self.columns, data=self.data)
        dt.refresh()
        self.assertEqual(len(dt), 3)

    def test_add_row_with_index(self):

        dt = DataTable(self.columns, data=self.data, index="a")
        dt.refresh()
        dt.add_row(dict(a=4, b=7.142, c="qux"))
        self.assertEqual(len(dt), 4)

    def test_add_row_without_index(self):

        dt = DataTable(self.columns, data=self.data)
        dt.refresh()
        dt.add_row(dict(a=4, b=7.142, c="qux"))
        self.assertEqual(len(dt), 4)



#!/usr/bin/env python3

import urwid

from .sparkwidgets import *

class ProgressBar(urwid.WidgetWrap):

    def __init__(self, width, maximum, value=0,
                 progress_color=None, remaining_color=None):
        self.width = width
        self.maximum = maximum
        self.value = value
        self.progress_color = progress_color or DEFAULT_BAR_COLOR
        self.remaining_color = remaining_color or DEFAULT_LABEL_COLOR
        self.placeholder = urwid.WidgetPlaceholder(urwid.Text(""))
        self.update()
        super().__init__(self.placeholder)

    def pack(self, size, focus=False):
        return (self.width, 1)

    @property
    def value_label(self):
        label_text = str(self.value)
        bar_len = self.spark_bar.bar_width(0)
        attr1 = f"{DEFAULT_LABEL_COLOR}:{self.progress_color}"
        content = [(attr1, label_text[:bar_len])]
        if len(label_text) > bar_len-1:
            attr2 = f"{DEFAULT_LABEL_COLOR}:{self.remaining_color}"
            content.append((attr2, label_text[bar_len:]))
        return urwid.Text(content)

    @property
    def maximum_label(self):
        label_text = str(self.maximum)
        bar_len = self.spark_bar.bar_width(1)
        attr1 = f"{DEFAULT_LABEL_COLOR}:{self.remaining_color}"
        content = []
        if bar_len:
            content.append((attr1, label_text[-bar_len:]))
        if len(label_text) > bar_len:
            attr2 = f"{DEFAULT_LABEL_COLOR}:{self.progress_color}"
            content.insert(0, (attr2, label_text[:-bar_len or None]))
        return urwid.Text(content)

    def update(self):
        value_label = None
        maximum_label = None

        self.spark_bar = SparkBarWidget(
            [
                SparkBarItem(self.value, bcolor=self.progress_color),
                SparkBarItem(self.maximum-self.value, bcolor=self.remaining_color),
            ], width=self.width
        )
        overlay1 = urwid.Overlay(
            urwid.Filler(self.value_label),
            urwid.Filler(self.spark_bar),
            "left",
            len(self.value_label.get_text()[0]),
            "top",
            1
        )
        label_len = len(self.maximum_label.get_text()[0])
        overlay2 =  urwid.Overlay(
            urwid.Filler(self.maximum_label),
            overlay1,
            "left",
            label_len,
            "top",
            1,
            left=self.width - label_len
        )
        self.placeholder.original_widget = urwid.BoxAdapter(overlay2, 1)

    def set_value(self, value):
        self.value = value
        self.update()

    @property
    def items(self):
        return self.spark_bar.items

__all__ = ["ProgressBar"]



from . import autocomplete
from .autocomplete import *
from . import datatable
from .datatable import *
from . import dialog
from .dialog import *
from . import dropdown
from .dropdown import *
from . import highlightable
from .highlightable import *
from . import keymap
from .keymap import *
from . import listbox
from .listbox import *
from . import scroll
from .scroll import *
from . import sparkwidgets
from .sparkwidgets import *
from . import tabview
from .tabview import *

__version__ = "0.3.4"

__all__ = (
    autocomplete.__all__
    + datatable.__all__
    + dialog.__all__
    + dropdown.__all__
    + highlightable.__all__
    + keymap.__all__
    + listbox.__all__
    + scroll.__all__
    + sparkwidgets.__all__
    + tabview.__all__
)



import logging
logger = logging.getLogger(__name__)
import itertools

import urwid

from .highlightable import HighlightableTextMixin
from .keymap import *

@keymapped()
class AutoCompleteEdit(urwid.Edit):

    signals = ["select", "close", "complete_next", "complete_prev"]

    KEYMAP = {
        "enter": "confirm",
        "esc": "cancel"
    }

    def clear(self):
        self.set_edit_text("")

    def confirm(self):
        self._emit("select")
        self._emit("close")

    def cancel(self):
        self._emit("close")

    def complete_next(self):
        self._emit("complete_next")

    def complete_prev(self):
        self._emit("complete_prev")

    def keypress(self, size, key):
        return super().keypress(size, key)

@keymapped()
class AutoCompleteBar(urwid.WidgetWrap):

    signals = ["change", "complete_prev", "complete_next", "select", "close"]

    prompt_attr = "dropdown_prompt"

    def __init__(self, prompt_attr=None):

        self.prompt_attr = prompt_attr or self.prompt_attr
        self.prompt = urwid.Text((self.prompt_attr, "> "))
        self.text = AutoCompleteEdit("")
        # self.text.selectable = lambda x: False
        self.cols = urwid.Columns([
            (2, self.prompt),
            ("weight", 1, self.text)
        ], dividechars=0)
        self.cols.focus_position = 1
        self.filler = urwid.Filler(self.cols, valign="bottom")
        urwid.connect_signal(self.text, "postchange", self.text_changed)
        urwid.connect_signal(self.text, "complete_prev", lambda source: self._emit("complete_prev"))
        urwid.connect_signal(self.text, "complete_next", lambda source: self._emit("complete_next"))
        urwid.connect_signal(self.text, "select", lambda source: self._emit("select"))
        urwid.connect_signal(self.text, "close", lambda source: self._emit("close"))
        super(AutoCompleteBar, self).__init__(self.filler)

    def set_prompt(self, text):

        self.prompt.set_text((self.prompt_attr, text))

    def set_text(self, text):

        self.text.set_edit_text(text)

    def text_changed(self, source, text):
        self._emit("change", text)

    def confirm(self):
        self._emit("select")
        self._emit("close")

    def cancel(self):
        self._emit("close")

    def __len__(self):
        return len(self.body)

    def keypress(self, size, key):
        return super().keypress(size, key)

@keymapped()
class AutoCompleteMixin(object):

    auto_complete = None
    prompt_attr = "dropdown_prompt"

    def __init__(self, auto_complete, prompt_attr=None, *args, **kwargs):
        super().__init__(self.complete_container, *args, **kwargs)
        if auto_complete is not None: self.auto_complete = auto_complete
        if prompt_attr is not None:
            self.prompt_attr = prompt_attr
        self.auto_complete_bar = None
        self.completing = False
        self.complete_anywhere = False
        self.case_sensitive = False
        self.last_complete_pos = None
        self.complete_string_location = None
        self.last_filter_text = None

        if self.auto_complete:
            self.auto_complete_bar = AutoCompleteBar(prompt_attr=self.prompt_attr)


            urwid.connect_signal(
                self.auto_complete_bar, "change",
                lambda source, text: self.complete()
            )
            urwid.connect_signal(
                self.auto_complete_bar, "complete_prev",
                lambda source: self.complete_prev()
            )
            urwid.connect_signal(
                self.auto_complete_bar, "complete_next",
                lambda source: self.complete_next()
            )

            urwid.connect_signal(
                self.auto_complete_bar, "select", self.on_complete_select
            )
            urwid.connect_signal(
                self.auto_complete_bar, "close", self.on_complete_close
            )

    def keypress(self, size, key):
        return super().keypress(size, key)
        # key = super().keypress(size, key)
        # if self.completing and key == "enter":
        #     self.on_complete_select(self)
        # else:
        #     return key

    @property
    def complete_container(self):
        raise NotImplementedError

    @property
    def complete_container_position(self):
        return 1

    @property
    def complete_body_position(self):
        return 0

    @property
    def complete_body(self):
        raise NotImplementedError

    @property
    def complete_items(self):
        raise NotImplementedError


    def complete_widget_at_pos(self, pos):
        return self.complete_body[pos]

    def complete_set_focus(self, pos):
        self.focus_position = pos

    @keymap_command()
    def complete_prefix(self):
        self.complete_on()

    @keymap_command()
    def complete_substring(self):
        self.complete_on(anywhere=True)

    def complete_prev(self):
        self.complete(step=-1)

    def complete_next(self):
        self.complete(step=1)

    def complete_on(self, anywhere=False, case_sensitive=False):

        if self.completing:
            return
        self.completing = True
        self.show_bar()
        if anywhere:
            self.complete_anywhere = True
        else:
            self.complete_anywhere = False

        if case_sensitive:
            self.case_sensitive = True
        else:
            self.case_sensitive = False

    def complete_compare_substring(self, search, candidate):
        try:
            return candidate.index(search)
        except ValueError:
            return None

    def complete_compare_fn(self, search, candidate):

        if self.case_sensitive:
            f = lambda x: str(x)
        else:
            f = lambda x: str(x.lower())

        if self.complete_anywhere:
            return self.complete_compare_substring(f(search), f(candidate))
        else:
            return 0 if self.complete_compare_substring(f(search), f(candidate))==0 else None
        # return f(candidate)


    @keymap_command()
    def complete_off(self):

        if not self.completing:
            return
        self.filter_text = ""

        self.hide_bar()
        self.completing = False

    @keymap_command
    def complete(self, step=None, no_wrap=False):

        if not self.filter_text:
            return

        # if not step and self.filter_text == self.last_filter_text:
        #     return

        logger.info(f"complete: {self.filter_text}")

        if self.last_complete_pos:
            widget = self.complete_widget_at_pos(self.last_complete_pos)
            if isinstance(widget, HighlightableTextMixin):
                widget.unhighlight()

        self.initial_pos = self.complete_body.get_focus()[1]
        positions = itertools.cycle(
            self.complete_body.positions(reverse=(step and step < 0))
        )
        pos = next(positions)
        while pos != self.initial_pos:
            logger.info(pos)
            pos = next(positions)
        for i in range(abs(step or 0)):
            pos = next(positions)

        while True:
            widget = self.complete_widget_at_pos(pos)
            complete_index = self.complete_compare_fn(self.filter_text, str(widget))
            if complete_index is not None:
                self.last_complete_pos = pos
                if isinstance(widget, HighlightableTextMixin):
                    widget.highlight(complete_index, complete_index+len(self.filter_text))
                self.complete_set_focus(pos)
                break
            pos = next(positions)
            if pos == self.initial_pos:
                break

        logger.info("done")
        self.last_filter_text = self.filter_text

    @keymap_command()
    def cancel(self):
        logger.debug("cancel")
        self.complete_container.focus_position = self.selected_button
        self.close()

    def close(self):
        self._emit("close")

    def show_bar(self):
        pos = self.complete_container_pos
        self.complete_container.contents[pos:pos+1] += [(
            self.auto_complete_bar,
            self.complete_container.options("given", 1)
        )]
        # self.box.height -= 1
        self.complete_container.focus_position = pos

    def hide_bar(self):
        pos = self.complete_container_pos
        widget = self.complete_widget_at_pos(self.complete_body.get_focus()[1])
        if isinstance(widget, HighlightableTextMixin):
            widget.unhighlight()
        self.complete_container.focus_position = self.complete_body_position
        del self.complete_container.contents[pos]
        # self.box.height += 1

    @property
    def filter_text(self):
        return self.auto_complete_bar.text.get_text()[0]

    @filter_text.setter
    def filter_text(self, value):
        return self.auto_complete_bar.set_text(value)

    def on_complete_select(self, source):
        widget = self.complete_widget_at_pos(self.complete_body.get_focus()[1])
        self.complete_off()
        self._emit("select", self.last_complete_pos, widget)
        self._emit("close")

    def on_complete_close(self, source):
        self.complete_off()

__all__ = ["AutoCompleteMixin"]



# Mixin class for mapping keyboard input to widget methods.

import logging
logger = logging.getLogger(__name__)

import six
import asyncio
import urwid
import re

KEYMAP_GLOBAL = {}

_camel_snake_re_1 = re.compile(r'(.)([A-Z][a-z]+)')
_camel_snake_re_2 = re.compile('([a-z0-9])([A-Z])')

def camel_to_snake(s):
    s = _camel_snake_re_1.sub(r'\1_\2', s)
    return _camel_snake_re_2.sub(r'\1_\2', s).lower()


def optional_arg_decorator(fn):
    def wrapped_decorator(*args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            return fn(args[0])
        else:
            def real_decorator(decoratee):
                return fn(decoratee, *args, **kwargs)
            return real_decorator
    return wrapped_decorator


@optional_arg_decorator
def keymap_command(f, command=None, *args, **kwargs):
    f._keymap = True
    f._keymap_command = command
    f._keymap_args = args
    f._keymap_kwargs = kwargs
    return f


def keymapped():

    def wrapper(cls):

        cls.KEYMAP_MERGED = {}

        if not hasattr(cls, "KEYMAP_SCOPE"):
            cls.KEYMAP_SCOPE = classmethod(lambda cls: camel_to_snake(cls.__name__))
        elif isinstance(cls.KEYMAP_SCOPE, str):
            cls.KEYMAP_SCOPE = classmethod(lambda cls: cls.KEYMAP_SCOPE)

        if not cls.KEYMAP_SCOPE() in cls.KEYMAP_MERGED:
            cls.KEYMAP_MERGED[cls.KEYMAP_SCOPE()] = {}
        if getattr(cls, "KEYMAP", False):
            cls.KEYMAP_MERGED[cls.KEYMAP_SCOPE()].update(**cls.KEYMAP)


        for base in cls.mro():
            if hasattr(base, "KEYMAP"):
                if not base.KEYMAP_SCOPE() in cls.KEYMAP_MERGED:
                    cls.KEYMAP_MERGED[base.KEYMAP_SCOPE()] = {}
                cls.KEYMAP_MERGED[base.KEYMAP_SCOPE()].update(**base.KEYMAP)

        # from pprint import pprint; print(cls.KEYMAP_MERGED)
        if not hasattr(cls, "KEYMAP_MAPPING"):
            cls.KEYMAP_MAPPING = {}

        cls.KEYMAP_MAPPING.update(**getattr(cls.__base__, "KEYMAP_MAPPING", {}))

        cls.KEYMAP_MAPPING.update({
            (getattr(getattr(cls, k), "_keymap_command", k) or k).replace(" ", "_"): k
            for k in cls.__dict__.keys()
            if hasattr(getattr(cls, k), '_keymap')
        })

        def keymap_command(self, cmd):
            logger.debug(f"keymap_command: {cmd}")
            args = []
            kwargs = {}

            if callable(cmd):
                f = cmd
            else:
                if isinstance(cmd, tuple):
                    if len(cmd) == 3:
                        (cmd, args, kwargs) = cmd
                    elif len(cmd) == 2:
                        if isinstance(cmd[1], dict):
                            (cmd, kwargs) = cmd
                        else:
                            (cmd, args) = cmd
                    else:
                        raise Exception
                elif isinstance(cmd, str):
                    cmd = cmd.replace(" ", "_")
                else:
                    logger.debug(f"keymap command {cmd} not valid")
                    return None

                if hasattr(self, cmd):
                    fn_name = cmd
                else:
                    try:
                        fn_name = self.KEYMAP_MAPPING[cmd]
                    except KeyError:
                        raise KeyError(cmd, self.KEYMAP_MAPPING, type(self))

                f = getattr(self, fn_name)

            ret = f(*args, **kwargs)
            if asyncio.iscoroutine(ret):
                asyncio.get_event_loop().create_task(ret)
            return None

        cls._keymap_command = keymap_command

        def keymap_register(self, key, cmd):
            self.KEYMAP_MERGED[cls.KEYMAP_SCOPE()][key] = cmd

        cls.keymap_register = keymap_register

        def keypress_decorator(func):


            def keypress(self, size, key):
                logger.debug(f"{cls} wrapped keypress: {key}, {cls.KEYMAP_SCOPE()}, {self.KEYMAP_MERGED.get(cls.KEYMAP_SCOPE(), {}).keys()}")

                if key and callable(func):
                    logger.debug(f"{cls} wrapped keypress, key: {key}, calling orig: {func}")
                    key = func(self, size, key)
                if key:
                    logger.debug(f"{cls} wrapped keypress, key: {key}, calling super: {super(cls, self).keypress}")
                    key = super(cls, self).keypress(size, key)
                keymap_combined = dict(self.KEYMAP_MERGED, **KEYMAP_GLOBAL)
                if key and keymap_combined.get(cls.KEYMAP_SCOPE(), {}).get(key, None):
                    cmd = keymap_combined[cls.KEYMAP_SCOPE()][key]
                    if isinstance(cmd, str) and cmd.startswith("keypress "):
                        new_key = cmd.replace("keypress ", "").strip()
                        logger.debug(f"{cls} remap {key} => {new_key}")
                        key = new_key
                    else:
                        logger.debug(f"{cls} wrapped keypress, key: {key}, calling keymap command")
                        key = self._keymap_command(cmd)
                return key

            return keypress

        cls.keypress = keypress_decorator(getattr(cls, "keypress", None))
        return cls

    return wrapper




@keymapped()
class KeymapMovementMixin(object):

    @classmethod
    def KEYMAP_SCOPE(cls):
        return "movement"

    def cycle_position(self, n):

        if len(self):
            pos = self.focus_position + n
            if pos > len(self) - 1:
                pos = len(self) - 1
            elif pos < 0:
                pos = 0
            self.focus_position = pos

    @keymap_command("up")
    def keymap_up(self): self.cycle_position(-1)

    @keymap_command("down")
    def keymap_down(self): self.cycle_position(1)

    @keymap_command("page up")
    def keymap_page_up(self): self.cycle_position(-self.page_size)

    @keymap_command("page down")
    def keymap_page_down(self): self.cycle_position(self.page_size)

    @keymap_command("home")
    def keymap_home(self): self.focus_position = 0

    @keymap_command("end")
    def keymap_end(self): self.focus_position = len(self)-1

__all__ = [
    "keymapped",
    "keymap_command",
    "KeymapMovementMixin"
]



import urwid
import itertools

DEFAULT_CELL_PADDING = 0

def partition(pred, iterable):
    'Use a predicate to partition entries into false entries and true entries'
    # partition(is_odd, range(10)) --> 0 2 4 6 8   and  1 3 5 7 9
    t1, t2 = itertools.tee(iterable)
    return itertools.filterfalse(pred, t1), filter(pred, t2)

intersperse = lambda e,l: sum([[x, e] for x in l],[])[:-1]

class DataTableText(urwid.Text):

    DEFAULT_END_CHAR = u"\N{HORIZONTAL ELLIPSIS}"

    def truncate(self, width, end_char=None):
        # raise Exception(width)
        text = self.get_text()[0]
        max_width = width
        if end_char:
            if end_char is True:
                end_char = self.DEFAULT_END_CHAR
            max_width -= len(end_char)
        else:
            end_char=""
        # raise Exception(len(text), max_width)
        if len(text) > max_width:
            self.set_text(text[:max_width]+end_char)

    def __len__(self):
        return len(self.get_text()[0])



"""
```sparkwidgets```
========================

A set of sparkline-ish widgets for urwid

This module contains a set of urwid text-like widgets for creating tiny but
hopefully useful sparkline-like visualizations of data.
"""

import urwid
from urwid_utils.palette import *
from collections import deque
import math
import operator
import itertools
import collections
from dataclasses import dataclass

BLOCK_VERTICAL = [ chr(x) for x in range(0x2581, 0x2589) ]
BLOCK_HORIZONTAL = [" "] + [ chr(x) for x in range(0x258F, 0x2587, -1) ]

DEFAULT_LABEL_COLOR = "black"
DEFAULT_LABEL_COLOR_DARK = "black"
DEFAULT_LABEL_COLOR_LIGHT = "white"

DEFAULT_BAR_COLOR = "white"

DISTINCT_COLORS_16 = urwid.display_common._BASIC_COLORS[1:]

DISTINCT_COLORS_256 = [
    '#f00', '#080', '#00f', '#d6f', '#0ad', '#f80', '#8f0', '#666',
    '#f88', '#808', '#0fd', '#66f', '#aa8', '#060', '#faf', '#860',
    '#60a', '#600', '#ff8', '#086', '#8a6', '#adf', '#88a', '#f60',
    '#068', '#a66', '#f0a', '#fda'
]

DISTINCT_COLORS_TRUE = [
    '#ff0000', '#008c00', '#0000ff', '#c34fff',
    '#01a5ca', '#ec9d00', '#76ff00', '#595354',
    '#ff7598', '#940073', '#00f3cc', '#4853ff',
    '#a6a19a', '#004301', '#edb7ff', '#8a6800',
    '#6100a3', '#5c0011', '#fff585', '#007b69',
    '#92b853', '#abd4ff', '#7e79a3', '#ff5401',
    '#0a577d', '#a8615c', '#e700b9', '#ffc3a6'
]

COLOR_SCHEMES = {
    "mono": {
        "mode": "mono"
    },
    "rotate_16": {
        "mode": "rotate",
        "colors": DISTINCT_COLORS_16
    },
    "rotate_256": {
        "mode": "rotate",
        "colors": DISTINCT_COLORS_256
    },
    "rotate_true": {
        "mode": "rotate",
        "colors": DISTINCT_COLORS_TRUE
    },
    "signed": {
        "mode": "rules",
        "colors": {
            "nonnegative": "default",
            "negative": "dark red"
        },
        "rules": [
            ( "<", 0, "negative" ),
            ( "else", "nonnegative" ),
        ]
    }
}

def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def get_palette_entries(
        chart_colors = None,
        label_colors = None
):

    NORMAL_FG_MONO = "white"
    NORMAL_FG_16 = "light gray"
    NORMAL_BG_16 = "black"
    NORMAL_FG_256 = "light gray"
    NORMAL_BG_256 = "black"

    palette_entries = {}

    if not label_colors:
        label_colors = list(set([
            DEFAULT_LABEL_COLOR,
            DEFAULT_LABEL_COLOR_DARK,
            DEFAULT_LABEL_COLOR_LIGHT
        ]))


    if chart_colors:
        colors = chart_colors
    else:
        colors = (urwid.display_common._BASIC_COLORS
                  + DISTINCT_COLORS_256
                  + DISTINCT_COLORS_TRUE )

    fcolors = colors + label_colors
    bcolors = colors

    for fcolor in fcolors:
        if isinstance(fcolor, PaletteEntry):
            fname = fcolor.name
            ffg = fcolor.foreground
            fbg = NORMAL_BG_16
            ffghi = fcolor.foreground_high
            fbghi = NORMAL_BG_256
        else:
            fname = fcolor
            ffg = (fcolor
                  if fcolor in urwid.display_common._BASIC_COLORS
                  else NORMAL_FG_16)
            fbg = NORMAL_BG_16
            ffghi = fcolor
            fbghi = NORMAL_BG_256

        palette_entries.update({
            fname: PaletteEntry(
                name = fname,
                mono = NORMAL_FG_MONO,
                foreground = ffg,
                background = fbg,
                foreground_high = ffghi,
                background_high = fbghi
            ),
        })

        for bcolor in bcolors:

            if isinstance(bcolor, PaletteEntry):
                bname = "%s:%s" %(fname, bcolor.name)
                bfg = ffg
                bbg = bcolor.background
                bfghi = ffghi
                bbghi = bcolor.background_high
            else:
                bname = "%s:%s" %(fname, bcolor)
                bfg = fcolor
                bbg = bcolor
                bfghi = fcolor
                bbghi = bcolor

            palette_entries.update({
                bname: PaletteEntry(
                    name = bname,
                    mono = NORMAL_FG_MONO,
                    foreground = (bfg
                                  if bfg in urwid.display_common._BASIC_COLORS
                                  else NORMAL_BG_16),
                    background = (bbg
                                  if bbg in urwid.display_common._BASIC_COLORS
                                  else NORMAL_BG_16),
                    foreground_high = bfghi,
                    background_high = bbghi
                ),
            })

    return palette_entries



OPERATOR_MAP = {
    "<": operator.lt,
    "<=": operator.le,
    ">": operator.gt,
    ">=": operator.ge,
    "=": operator.eq,
    "else": lambda a, b: True
}


class SparkWidget(urwid.Text):

    @staticmethod
    def make_rule_function(scheme):

        rules = scheme["rules"]
        def rule_function(value):
            return scheme["colors"].get(
                next(iter(filter(
                    lambda rule: all(
                        OPERATOR_MAP[cond[0]](value, cond[1] if len(cond) > 1 else None)
                        for cond in [rule]
                    ), scheme["rules"]
                )))[-1]
            )
        return rule_function


    @staticmethod
    def normalize(v, a, b, scale_min, scale_max):

        if scale_max == scale_min:
            return v
        return max(
            a,
            min(
                b,
                (((v - scale_min) / (scale_max - scale_min) ) * (b - a) + a)
            )
        )


    def parse_scheme(self, scheme):

        if isinstance(scheme, dict):
            color_scheme = scheme
        else:
            try:
                color_scheme = COLOR_SCHEMES[scheme]
            except:
                return lambda x: scheme
                # raise Exception("Unknown color scheme: %s" %(scheme))

        mode = color_scheme["mode"]
        if mode == "mono":
            return None

        elif mode == "rotate":
            return deque(color_scheme["colors"])

        elif mode == "rules":
            return self.make_rule_function(color_scheme)

        else:
            raise Exception("Unknown color scheme mode: %s" %(mode))

    @property
    def current_color(self):

        return self.colors[0]

    def next_color(self):
        if not self.colors:
            return
        self.colors.rotate(-1)
        return self.current_color

    def get_color(self, item):
        if not self.colors:
            color = None
        elif callable(self.colors):
            color = self.colors(item)
        elif isinstance(self.colors, collections.Iterable):
            color = self.current_color
            self.next_color()
            return color
        else:
            raise Exception(self.colors)

        return color


class SparkColumnWidget(SparkWidget):
    """
    A sparkline-ish column widget for Urwid.

    Given a list of numeric values, this widget will draw a small text-based
    vertical bar graph of the values, one character per value.  Column segments
    can be colorized according to a color scheme or by assigning each
    value a color.

    :param items: A list of items to be charted in the widget.  Items can be
    either numeric values or tuples, the latter of which must be of the form
    ('attribute', value) where attribute is an urwid text attribute and value
    is a numeric value.

    :param color_scheme: A string or dictionary containing the name of or
    definition of a color scheme for the widget.

    :param underline: one of None, "negative", or "min", specifying values that
    should be marked in the chart.  "negative" shows negative values as little
    dots at the bottom of the chart, while "min" uses a unicode combining
    three dots character to indicate minimum values.  Results of this and the
    rest of these parameters may not look great with all terminals / fonts,
    so if this looks weird, don't use it.

    :param overline: one of None or "max" specfying values that should be marked
    in the chart.  "max" draws three dots above the max value.  See underline
    description for caveats.

    :param scale_min: Set a minimum scale for the chart.  By default, the range
    of the chart's Y axis will expand to show all values, but this parameter
    can be used to restrict or expand the Y-axis.

    :param scale_max: Set the maximum for the Y axis. -- see scale_min.
    """

    chars = BLOCK_VERTICAL

    def __init__(self, items,
                 color_scheme = "mono",
                 scale_min = None,
                 scale_max = None,
                 underline = None,
                 overline = None,
                 *args, **kwargs):

        self.items = items
        self.colors = self.parse_scheme(color_scheme)

        self.underline = underline
        self.overline = overline

        self.values = [ i[1] if isinstance(i, tuple) else i for i in self.items ]

        v_min = min(self.values)
        v_max = max(self.values)


        def item_to_glyph(item):

            color = None

            if isinstance(item, tuple):
                color = item[0]
                value = item[1]
            else:
                color = self.get_color(item)
                value = item

            if self.underline == "negative" and value < 0:
                glyph = " \N{COMBINING DOT BELOW}"
            else:


                # idx = scale_value(value, scale_min=scale_min, scale_max=scale_max)
                idx = self.normalize(
                    value, 0, len(self.chars)-1,
                    scale_min if scale_min else v_min,
                    scale_max if scale_max else v_max)

                glyph = self.chars[int(round(idx))]

                if self.underline == "min" and value == v_min:
                    glyph = "%s\N{COMBINING TRIPLE UNDERDOT}" %(glyph)

                if self.overline == "max" and value == v_max:
                    glyph = "%s\N{COMBINING THREE DOTS ABOVE}" %(glyph)

            if color:
                return (color, glyph)
            else:
                return glyph

        self.sparktext = [
            item_to_glyph(i)
            for i in self.items
        ]
        super(SparkColumnWidget, self).__init__(self.sparktext, *args, **kwargs)


# via https://github.com/rg3/dhondt
def dhondt_formula(votes, seats):
    return votes / (seats + 1)

def bar_widths(party_votes, total_seats):
    # Calculate the quotients matrix (list in this case).
    quot = []
    ret = dict()
    for p in dict(enumerate(party_votes)):
        ret[p] = 0
        for s in range(0, total_seats):
            q = dhondt_formula(party_votes[p], s)
            quot.append((q, p))

    # Sort the quotients by value.
    quot.sort(reverse=True)

    # Take the highest quotients with the assigned parties.
    for s in range(0, total_seats):
        ret[quot[s][1]] += 1
    return list(ret.values())


@dataclass
class SparkBarItem:

    value: int
    label: str = None
    fcolor: str = None
    bcolor: str = None
    align: str = "<"
    fill: str = " "

    @property
    def steps(self):
        return len(BLOCK_HORIZONTAL)

    def formatted_label(self, total):
        if self.label is None:
            return None
        try:
            pct = int(round(self.value/total*100, 0))
        except:
            pct = ""

        return str(self.label).format(
            value=self.value,
            pct=pct
        )
    def truncated_label(self, width, total):

        label = self.formatted_label(total)
        if not label:
            return None
        return (
            label[:width-1] + "\N{HORIZONTAL ELLIPSIS}"
            if len(label) > width
            else label
        )

        # s = "{label:.{n}}".format(
        #     label=self.formatted_label(total),
        #     n=min(len(label), width),
        # )
        # if len(s) > width:
        #     chars[-1] = "\N{HORIZONTAL ELLIPSIS}"


    def output(self, width, total, next_color=None):

        steps_width = width % self.steps if next_color else None
        chars_width = width // self.steps# - (1 if steps_width else 0)
        # print(width, chars_width, steps_width)
        label = self.truncated_label(chars_width, total)
        if label:
            chars = "{:{a}{m}.{m}}".format(
                label,
                m=max(chars_width, 0),
                a=self.align or "<",
            )
            # if len(label) > chars_width:
            #     chars[-1] = "\N{HORIZONTAL ELLIPSIS}"
        else:
            chars = self.fill * chars_width



        output = [
            (
                "%s:%s" %(
                    self.fcolor or DEFAULT_LABEL_COLOR,
                    self.bcolor or DEFAULT_BAR_COLOR), chars
            )
        ]

        if steps_width:
            attr = f"{self.bcolor}:{next_color}"
            output.append(
                (attr, BLOCK_HORIZONTAL[steps_width])
            )
        return output


class SparkBarWidget(SparkWidget):
    """
    A sparkline-ish horizontal stacked bar widget for Urwid.

    This widget graphs a set of values in a horizontal bar style.

    :param items: A list of items to be charted in the widget.  Items can be
    either numeric values or tuples, the latter of which must be of the form
    ('attribute', value) where attribute is an urwid text attribute and value
    is a numeric value.

    :param width: Width of the widget in characters.

    :param color_scheme: A string or dictionary containing the name of or
    definition of a color scheme for the widget.
    """

    fill_char = " "

    def __init__(self, items, width,
                 color_scheme="mono",
                 label_color=None,
                 min_width=None,
                 fit_label=False,
                 normalize=None,
                 fill_char=None,
                 *args, **kwargs):

        self.items = [
            i if isinstance(i, SparkBarItem) else SparkBarItem(i)
            for i in items
        ]

        self.colors = self.parse_scheme(color_scheme)

        for i in self.items:
            if not i.bcolor:
                i.bcolor = self.get_color(i)
            if fill_char:
                i.fill_char = fill_char

        self.width = width
        self.label_color = label_color
        self.min_width = min_width
        self.fit_label = fit_label

        values = None
        total = None

        if normalize:
            values = [ item.value for item in self.items ]
            v_min = min(values)
            v_max = max(values)
            values = [
                int(self.normalize(v,
                                   normalize[0], normalize[1],
                                   v_min, v_max))
                for v in values
            ]
            for i, v in enumerate(values):
                self.items[i].value = v


        filtered_items = self.items
        values = [i.value for i in filtered_items]
        total = sum(values)

        charwidth = total / self.width

        self.sparktext = []

        position = 0
        lastcolor = None

        values = [i.value for i in filtered_items]

        # number of steps that can be represented within each screen character
        # represented by Unicode block characters
        steps = len(BLOCK_HORIZONTAL)

        # use a prorportional representation algorithm to distribute the number
        # of available steps among each bar segment
        self.bars = bar_widths(values, self.width*steps)

        if self.min_width or self.fit_label:
            # make any requested adjustments to bar widths
            for i in range(len(self.bars)):
                if self.min_width and self.bars[i] < self.min_width*steps:
                    self.bars[i] = self.min_width*steps
                if self.fit_label:
                    # need some slack here to compensate for self.bars that don't
                    # begin on a character boundary
                    label_len = len(self.items[i].formatted_label(total))+2
                    if self.bars[i] < label_len*steps:
                        self.bars[i] = label_len*steps
            # use modified proportions to calculate new proportions that try
            # to account for min_width and fit_label
            self.bars = bar_widths(self.bars, self.width*steps)

        # filtered_items = [item for i, item in enumerate(self.items) if self.bars[i]]
        # self.bars = [b for b in self.bars if b]

        for i, (item, item_next) in enumerate(pairwise(filtered_items)):
            width = self.bars[i]
            output = item.output(width, total=total, next_color=item_next.bcolor)
            self.sparktext += output

        output = filtered_items[-1].output(self.bars[-1], total=total)
        self.sparktext += output

        if not self.sparktext:
            self.sparktext = ""
        self.set_text(self.sparktext)
        super(SparkBarWidget, self).__init__(self.sparktext, *args, **kwargs)

    def bar_width(self, index):
        return self.bars[index]//len(BLOCK_HORIZONTAL)


__all__ = [
    "SparkColumnWidget", "SparkBarWidget", "SparkBarItem",
    "get_palette_entries",
    "DEFAULT_LABEL_COLOR", "DEFAULT_LABEL_COLOR_DARK", "DEFAULT_LABEL_COLOR_LIGHT"
]



import logging
logger = logging.getLogger(__name__)
import os
import random
import string
from functools import wraps
import re
import itertools

import six
import urwid
from urwid_utils.palette import *
# import urwid_readline
from orderedattrdict import AttrDict

# from .datatable import *
from .listbox import ScrollingListBox
from .keymap import *
from .highlightable import HighlightableTextMixin
from .autocomplete import AutoCompleteMixin

class DropdownButton(urwid.Button):

    text_attr = "dropdown_text"

    left_chars = u""
    right_chars = u""


    def __init__(
            self, label,
            text_attr=None,
            left_chars=None, right_chars=None
    ):

        self.label_text = label
        if text_attr:
            self.text_attr = text_attr
        if left_chars:
            self.left_chars = left_chars
        if right_chars:
            self.right_chars = right_chars

        self.button_left = urwid.Text(self.left_chars)
        self.button_right = urwid.Text(self.right_chars)

        self._label = urwid.SelectableIcon("", cursor_position=0)
        self.cols = urwid.Columns([
            (len(self.left_chars), self.button_left),
            ('weight', 1, self._label),
            (len(self.right_chars), self.button_right)
        ], dividechars=0)
        self.set_label((self.text_attr, self.label_text))
        super(urwid.Button, self).__init__(self.cols)

    @property
    def decoration_width(self):
        return len(self.left_chars) + len(self.right_chars)

    @property
    def width(self):
        return self.decoration_width + len(self.label_text)


class DropdownItem(HighlightableTextMixin, urwid.WidgetWrap):

    signals = ["click"]

    text_attr = "dropdown_text"
    highlight_attr = "dropdown_highlight"
    focused_attr = "dropdown_focused"

    def __init__(self, label, value,
                 margin=0,
                 text_attr=None,
                 focused_attr=None,
                 highlight_attr=None,
                 left_chars=None, right_chars=None):

        self.label_text = label
        self.value = value
        self.margin = margin
        if text_attr:
            self.text_attr = text_attr
        if focused_attr:
            self.focused_attr = focused_attr
        if highlight_attr:
            self.highlight_attr = highlight_attr
        self.button = DropdownButton(
            self.label_text,
            text_attr=self.text_attr,
            left_chars=left_chars, right_chars=right_chars
        )

        self.padding = urwid.Padding(self.button, width=("relative", 100),
                                     left=self.margin, right=self.margin)


        self.attr = urwid.AttrMap(self.padding, {None: self.text_attr})
        self.attr.set_focus_map({
            None: self.focused_attr,
            self.text_attr: self.focused_attr
        })
        super(DropdownItem, self).__init__(self.attr)
        urwid.connect_signal(
            self.button,
            "click",
            lambda source: self._emit("click")
        )

    @property
    def highlight_source(self):
        return self.label_text

    @property
    def highlightable_attr_normal(self):
        return self.text_attr

    @property
    def highlightable_attr_highlight(self):
        return self.highlight_attr

    def on_highlight(self):
        self.set_text(self.highlight_content)

    def on_unhighlight(self):
        self.set_text(self.highlight_source)

    @property
    def width(self):
        return self.button.width + 2*self.margin

    @property
    def decoration_width(self):
        return self.button.decoration_width + 2*self.margin

    def __str__(self):
        return self.label_text

    def __contains__(self, s):
        return s in self.label_text

    def startswith(self, s):
        return self.label_text.startswith(s)

    @property
    def label(self):
        return self.button.label

    def set_text(self, text):
        self.button.set_label(text)

@keymapped()
class DropdownDialog(AutoCompleteMixin, urwid.WidgetWrap, KeymapMovementMixin):

    signals = ["select", "close"]

    text_attr = "dropdown_text"

    min_width = 4

    label = None
    border = None
    scrollbar = False
    margin = 0
    max_height = None

    def __init__(
            self,
            drop_down,
            items,
            default=None,
            label=None,
            border=False,
            margin = None,
            scrollbar=None,
            text_attr=None,
            focused_attr=None,
            prompt_attr=None,
            left_chars=None,
            right_chars=None,
            left_chars_top=None,
            rigth_chars_top=None,
            max_height=None,
            keymap = {},
            **kwargs
    ):

        self.drop_down = drop_down
        self.items = items
        if label is not None: self.label = label
        if border is not None: self.border = border
        if margin is not None: self.margin = margin
        if scrollbar is not None: self.scrollbar = scrollbar
        if text_attr:
            self.text_attr = text_attr
        if focused_attr:
            self.focused_attr = focused_attr
        if prompt_attr:
            self.prompt_attr = prompt_attr
        if max_height is not None: self.max_height = max_height
        self.selected_button = 0
        buttons = []

        buttons = [
                DropdownItem(
                    label=l, value=v, margin=self.margin,
                    text_attr=self.text_attr,
                    focused_attr=self.focused_attr,
                    left_chars=left_chars,
                    right_chars=right_chars,
                )
                for l, v in self.items.items()
        ]
        self.dropdown_buttons = ScrollingListBox(
            urwid.SimpleListWalker(buttons), with_scrollbar=scrollbar
        )

        urwid.connect_signal(
            self.dropdown_buttons,
            'select',
            lambda source, selection: self.on_complete_select(source)
        )

        kwargs = {}
        if self.label is not None:
            kwargs["title"] = self.label
            kwargs["tlcorner"] = u"\N{BOX DRAWINGS LIGHT DOWN AND HORIZONTAL}"
            kwargs["trcorner"] = u"\N{BOX DRAWINGS LIGHT DOWN AND LEFT}"

        w = self.dropdown_buttons
        if self.border:
           w = urwid.LineBox(w, **kwargs)

        self.pile = urwid.Pile([
            ("weight", 1, w),
        ])
        super().__init__(self.pile)

    @property
    def complete_container(self):
        return self.pile

    @property
    def complete_container_pos(self):
        return 1

    @property
    def complete_body(self):
        return self.body

    @property
    def complete_items(self):
        return self.body

    @property
    def max_item_width(self):
        if not len(self):
            return self.min_width
        return max(w.width for w in self)

    @property
    def width(self):
        width = self.max_item_width
        if self.border:
            width += 2
        return width

    @property
    def height(self):
        height = min(len(self), self.max_height)
        if self.border:
            height += 2
        return height

    @property
    def body(self):
        return self.dropdown_buttons.body

    def __getitem__(self, i):
        return self.body[i]

    def __len__(self):
        return len(self.body)

    @property
    def focus_position(self):
        return self.dropdown_buttons.focus_position

    @focus_position.setter
    def focus_position(self, pos):
        self.dropdown_buttons.listbox.set_focus_valign("top")
        self.dropdown_buttons.focus_position = pos

    @property
    def selection(self):
        return self.dropdown_buttons.selection

    # def on_complete_select(self, pos, widget):

    #     # logger.debug("select_button: %s" %(button))
    #     label = widget.label
    #     value = widget.value
    #     self.selected_button = self.focus_position
    #     self.complete_off()
    #     self._emit("select", widget)
    #     self._emit("close")

    # def keypress(self, size, key):
    #     return super(DropdownDialog, self).keypress(size, key)


    @property
    def selected_value(self):
        if not self.focus_position:
            return None
        return self.body[self.focus_position].value

@keymapped()
class Dropdown(urwid.PopUpLauncher):
    # Based in part on SelectOne widget from
    # https://github.com/tuffy/python-audio-tools

    signals = ["change"]

    text_attr = "dropdown_text"
    label_attr = "dropdown_label"
    focused_attr = "dropdown_focused"
    highlight_attr = "dropdown_highlight"
    prompt_attr = "dropdown_prompt"

    auto_complete = None
    label = None
    empty_label = u"\N{EMPTY SET}"
    expanded = False
    margin = 0

    def __init__(
            self,
            items=None,
            label=None,
            default=None,
            expanded=None,
            border=False, scrollbar=False,
            margin=None,
            text_attr=None,
            label_attr=None,
            focused_attr=None,
            highlight_attr=None,
            prompt_attr=None,
            left_chars=None, right_chars=None,
            left_chars_top=None, right_chars_top=None,
            auto_complete=None,
            max_height=10,
            # keymap = {}
    ):

        if items is not None:
            self._items = items
        if label is not None:
            self.label = label
        if expanded is not None:
            self.expanded = expanded
        self.default = default

        self.border = border
        self.scrollbar = scrollbar
        if auto_complete is not None: self.auto_complete = auto_complete

        # self.keymap = keymap

        if margin:
            self.margin = margin

        if text_attr:
            self.text_attr = text_attr
        if label_attr:
            self.label_attr = label_attr
        if focused_attr:
            self.focused_attr = focused_attr
        if highlight_attr:
            self.highlight_attr = highlight_attr
        if prompt_attr:
            self.prompt_attr = prompt_attr

        if isinstance(self.items, list):
            if len(self.items):
                if isinstance(self.items[0], tuple):
                    self._items = AttrDict(self.items)
                else:
                    logger.debug(self.items)
                    self._items = AttrDict(( (item, n) for n, item in enumerate(self.items)))
            else:
                self._items = AttrDict()
        else:
            self._items = self.items


        self.button = DropdownItem(
            u"", None,
            margin=self.margin,
            text_attr=self.text_attr,
            highlight_attr=self.highlight_attr,
            focused_attr=self.focused_attr,
            left_chars = left_chars_top if left_chars_top else left_chars,
            right_chars = right_chars_top if right_chars_top else right_chars
        )

        self.pop_up = DropdownDialog(
            self,
            self._items,
            self.default,
            label=self.label,
            border=self.border,
            margin=self.margin,
            text_attr=self.text_attr,
            focused_attr=self.focused_attr,
            prompt_attr=self.prompt_attr,
            left_chars=left_chars,
            right_chars=right_chars,
            auto_complete=self.auto_complete,
            scrollbar=scrollbar,
            max_height=max_height,
            # keymap=self.KEYMAP
        )

        urwid.connect_signal(
            self.pop_up,
            "select",
            lambda souce, pos, selection: self.select(selection)
        )

        urwid.connect_signal(
            self.pop_up,
            "close",
            lambda source: self.close_pop_up()
        )

        if self.default is not None:
            try:
                if isinstance(self.default, str):
                    try:
                        self.select_label(self.default)
                    except ValueError:
                        pass
                else:
                    raise StopIteration
            except StopIteration:
                try:
                    self.select_value(self.default)
                except StopIteration:
                    self.focus_position = 0

        if len(self):
            self.select(self.selection)
        else:
            self.button.set_text((self.text_attr, self.empty_label))

        cols = [ (self.button_width, self.button) ]

        if self.label:
            cols[0:0] = [
                ("pack", urwid.Text([(self.label_attr, "%s: " %(self.label))])),
            ]
        self.columns = urwid.Columns(cols, dividechars=0)

        w = self.columns
        if self.border:
            w = urwid.LineBox(self.columns)
        w = urwid.Padding(w, width=self.width)

        super(Dropdown, self).__init__(w)
        urwid.connect_signal(
            self.button,
            'click',
            lambda button: self.open_pop_up()
        )
        if self.expanded:
            self.open_pop_up()

    @classmethod
    def get_palette_entries(cls):
        return {
            "dropdown_text": PaletteEntry(
                foreground="light gray",
                background="dark blue",
                foreground_high="light gray",
                background_high="#003",
            ),
            "dropdown_focused": PaletteEntry(
                foreground="white",
                background="light blue",
                foreground_high="white",
                background_high="#009",
            ),
            "dropdown_highlight": PaletteEntry(
                foreground="yellow",
                background="light blue",
                foreground_high="yellow",
                background_high="#009",
            ),
            "dropdown_label": PaletteEntry(
                foreground="white",
                background="black"
            ),
            "dropdown_prompt": PaletteEntry(
                foreground="light blue",
                background="black"
            )
        }


    @keymap_command()
    def complete_prefix(self):
        if not self.auto_complete:
            return
        self.open_pop_up()
        self.pop_up.complete_prefix()

    @keymap_command()
    def complete_substring(self):
        if not self.auto_complete:
            return
        self.open_pop_up()
        self.pop_up.complete_substring()

    def create_pop_up(self):
        # print("create")
        return self.pop_up

    @property
    def button_width(self):
        return self.pop_up.max_item_width + self.button.decoration_width

    @property
    def pop_up_width(self):
        w = self.button_width
        if self.border:
            w += 2
        return w

    @property
    def contents_width(self):
        # raise Exception(self.button.width)
        w = self.button_width
        if self.label:
            w += len(self.label) + 2
        return max(self.pop_up.width, w)

    @property
    def width(self):
        width = max(self.contents_width, self.pop_up.width)
        if self.border:
            width += 2
        return width

    @property
    def height(self):
        height = self.pop_up.height + 1
        return height

    def pack(self, size, focus=False):
        return (self.width, self.height)

    @property
    def page_size(self):
        return self.pop_up.height

    def open_pop_up(self):
        # print("open")
        super(Dropdown, self).open_pop_up()

    def close_pop_up(self):
        super().close_pop_up()

    def get_pop_up_parameters(self):
        return {'left': (len(self.label) + 2 if self.label else 0),
                'top': 0,
                'overlay_width': self.pop_up_width,
                'overlay_height': self.pop_up.height
        }

    @property
    def focus_position(self):
        return self.pop_up.focus_position

    @focus_position.setter
    def focus_position(self, pos):
        if pos == self.focus_position:
            return
        # self.select_index(pos)
        old_pos = self.focus_position
        self.pop_up.selected_button = self.pop_up.focus_position = pos
        self.select(self.selection)

    @property
    def items(self):
        return self._items

    @property
    def selection(self):
        return self.pop_up.selection

    def select_label(self, label, case_sensitive=False):

        old_value = self.value

        f = lambda x: x
        if not case_sensitive:
            f = lambda x: x.lower()

        index = next(itertools.dropwhile(
                lambda x: f(x[1]) != f(label),
                enumerate((self._items.keys())
            )
        ))[0]
        self.focus_position = index



    @property
    def items(self):
        return self._items

    @property
    def selection(self):
        return self.pop_up.selection


    def select_label(self, label, case_sensitive=False):

        old_value = self.value

        f = lambda x: x
        if not case_sensitive:
            f = lambda x: x.lower() if isinstance(x, str) else x

        try:
            index = next(itertools.dropwhile(
                    lambda x: f(x[1]) != f(label),
                    enumerate((self._items.keys())
                )
            ))[0]
        except StopIteration:
            raise ValueError
        self.focus_position = index



    def select_value(self, value):

        index = next(itertools.dropwhile(
                lambda x: x[1] != value,
                enumerate((self._items.values())
            )
        ))[0]
        self.focus_position = index


    @property
    def labels(self):
        return self._items.keys()

    @property
    def values(self):
        return self._items.values()

    @property
    def selected_label(self):
        return self.selection.label

    @selected_label.setter
    def selected_label(self, label):
        return self.select_label(label)

    @property
    def selected_value(self):
        if not self.selection:
            return None
        return self.selection.value

    @selected_value.setter
    def selected_value(self, value):
        return self.select_value(value)

    @property
    def value(self):
        return self.selected_value

    @value.setter
    def value(self, value):
        old_value = self.value

        # try to set by value.  if not found, try to set by label
        try:
            self.selected_value = value
        except StopIteration:
            self.selected_label = value

    def cycle_prev(self):
        self.cycle(-1)

    def action(self):
        pass

    @keymap_command("cycle")
    def cycle(self, n):
        pos = self.focus_position + n
        if pos > len(self) - 1:
            pos = len(self) - 1
        elif pos < 0:
            pos = 0
        # self.focus_position = pos
        self.focus_position = pos

    def select(self, button):
        logger.debug("select: %s" %(button))
        self.button.set_text((self.text_attr, button.label))
        self.pop_up.dropdown_buttons.listbox.set_focus_valign("top")
        # if old_pos != pos:
        self.action()
        self._emit("change", self.selected_label, self.selected_value)

    # def set_items(self, items, selected_value):
    #     self._items = items
    #     self.make_selection([label for (label, value) in items if
    #                          value is selected_value][0],
    #                         selected_value)
    def __len__(self):
        return len(self.items)

__all__ = ["Dropdown"]



import functools
import itertools

import urwid

from .cells import *
from .columns import *
from orderedattrdict import AttrDict

class DataTableRow(urwid.WidgetWrap):

    def __init__(self, table,
                 content=None,
                 row_height=None,
                 divider=None, padding=None,
                 cell_selection=False,
                 style = None,
                 *args, **kwargs):

        self.table = table
        self.row_height = row_height
        self.content = content
        # if not isinstance(self.content, int):
        #     raise Exception(self.content, type(self))
        self.divider = divider
        self.padding = padding
        self.cell_selection = cell_selection
        self.style = style
        # self.details = None
        self.sort = self.table.sort_by
        self.attr = self.ATTR
        self.attr_focused = "%s focused" %(self.attr)
        self.attr_column_focused = "%s column_focused" %(self.attr)
        self.attr_highlight = "%s highlight" %(self.attr)
        self.attr_highlight_focused = "%s focused" %(self.attr_highlight)
        self.attr_highlight_column_focused = "%s column_focused" %(self.attr_highlight)
        self.attr_map =  {
            None: self.attr,
        }

        self.focus_map = {
            self.attr: self.attr_focused,
            self.attr_highlight: self.attr_highlight_focused,
        }

        # needed to restore if cell selection is toggled
        self.original_focus_map = self.focus_map.copy()

        # if self.cell_selection:
        self.focus_map.update({
            self.attr_focused: self.attr_column_focused,
            self.attr_highlight_focused: self.attr_highlight_column_focused,
        })
        self.focus_map.update(self.table.column_focus_map)
        self.cell_selection_focus_map = self.focus_map.copy()

        if cell_selection:
            self.enable_cell_selection()
        else:
            self.disable_cell_selection()

        self.focus_map.update(table.focus_map)

        self.contents_placeholder = urwid.WidgetPlaceholder(urwid.Text(""))

        w = self.contents_placeholder

        self.update()

        # if self.row_height:
        self.box = urwid.BoxAdapter(w, self.row_height or 1)

        self.pile = urwid.Pile([
            ("weight", 1, self.box)
        ])

        self.attrmap = urwid.AttrMap(
            self.pile,
            attr_map = self.attr_map,
            focus_map = self.focus_map,
        )

        super(DataTableRow, self).__init__(self.attrmap)

    def on_resize(self):

        if self.row_height is not None:
            return
        l = [1]
        # for i, c in enumerate(self.cells):
        for c, w in zip(self.cells, self.column_widths( (self.table.width,) )):
            # try:
            # c.contents.render( (self.table.visible_columns[i].width,), False)
            # except Exception as e:
            #     raise Exception(c, c.contents, e)

            try:
                rows = c.contents.rows( (w,) )
            except AttributeError:
                continue
            # logger.debug(f"{c}, {c.contents}, {w}, {rows}")

            # self.table.header.render((self.table.width, self.row_height), False)
            # raise Exception(self.table.header.data_cells[i].width)
            # rows = self.table.header.rows( (self.table.header.cells[i].width,) )
            # except Exception as e:
                # raise Exception(type(self), type(self.contents), e)
            # print(c, rows)
            l.append(rows)
        self.box.height = max(l)

        if self.details_open:
            self.open_details()

        # logger.debug(f"height: {self.box.height}")
        # (w, o) = self.pile.contents[0]
        # self.pile.contents[0] = (w, self.pile.options("given", max(l)))


    def keypress(self, size, key):
        try:
            key = super(DataTableRow, self).keypress(size, key)
        except AttributeError:
            pass
        return key

    def enable_cell_selection(self):
        self.cell_selection = True
        self.focus_map = self.cell_selection_focus_map

    def disable_cell_selection(self):
        self.cell_selection = False
        self.focus_map = self.original_focus_map

    def resize_column(self, index, width):
        # col = self.table.visible_columns[index*2]
        (widget, options) = self.columns.contents[index*2]
        self.columns.contents[index*2] = (widget, self.columns.options(*width))

    def make_columns(self):

        # logger.info("make_columns")
        self.cells = self.make_cells()

        columns = urwid.Columns([])

        idx = None
        for i, cell in enumerate(self.cells):
            if not (idx or isinstance(cell, DataTableDividerCell)):
                idx = i
            col = self.table.visible_columns[i]
            options = columns.options(col.sizing, col.width_with_padding(self.padding))
            columns.contents.append(
                (cell, options)

            )
        if idx:
            columns.focus_position = idx
        return columns

    def make_contents(self):
        self.columns = self.make_columns()
        return self.columns

    @property
    def contents(self):
        return self.contents_placeholder.original_widget

    def update(self):
        contents = self.make_contents()
        # if self.row_height is None:
        #     contents = urwid.Filler(contents)
        self.contents_placeholder.original_widget = contents

    def selectable(self):
        return True

    def set_focus_column(self, index):
        for i, cell in enumerate(self):
            if i == index:
                cell.highlight()
            else:
                cell.unhighlight()
    def __len__(self):
        return len(self.columns.contents)

    def __iter__(self):
        return iter( self.columns[i] for i in range(0, len(self.columns.contents)) )

    @property
    def values(self):
        return AttrDict(list(zip([c.name for c in self.table.visible_columns], [ c.value for c in self ])))

    @property
    def data_cells(self):
        return [ c for c in self.cells if not isinstance(c, DataTableDividerCell)]

    def column_widths(self, size=None):
        if not size:
            size = (self.table.width,)
        return self.columns.column_widths(size)

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height


class DataTableDetails(urwid.WidgetWrap):

    def __init__(self, row, content, indent=None):

        self.row = row
        self.contents = content
        self.columns = urwid.Columns([
            ("weight", 1, content)
        ])
        if indent:
            self.columns.contents.insert(0,
                (urwid.Padding(urwid.Text(" ")),
                 self.columns.options("given", indent)
                )
            )

        super().__init__(self.columns)

    def selectable(self):
        return not self.row.details_disabled


class DataTableBodyRow(DataTableRow):

    ATTR = "table_row_body"

    DIVIDER_CLASS = DataTableDividerBodyCell


    @property
    def index(self):
        return self.content

    @property
    def data(self):
        return AttrDict(self.table.get_dataframe_row(self.index))

    @property
    def data_source(self):
        return self.table.get_dataframe_row_object(self.index)

    def __getitem__(self, column):
        cls = self.table.df[self.index, "_cls"]
        # row = self.data
        if column in self.table.df.columns:
            # logger.info(f"__getitem__: {column}={self.table.df.get(self.index, column)}")
            return self.table.df[self.index, column]
        else:
            raise KeyError
            # raise Exception(column, self.table.df.columns)


    def __setitem__(self, column, value):
        self.table.df[self.index, column] = value
        # logger.info(f"__setitem__: {column}, {value}, {self.table.df[self.index, column]}")

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    @property
    def details_open(self):
        # logger.info(f"{self['_details']}")
        # raise Exception(self.get([self.index, "_details"], {}))
        return self.get("_details", {}).get("open", False)

    @details_open.setter
    def details_open(self, value):
        details = self["_details"]
        details["open"] = value
        self["_details"] = details

    @property
    def details_disabled(self):
        return (not self.table.detail_selectable) or self.get([self.index, "_details"], {}).get("disabled", False)

    @details_disabled.setter
    def details_disabled(self, value):
        details = self["_details"]
        details["disabled"] = value
        if value == True:
            self.details_focused = False
        self["_details"] = details

    @property
    def details_focused(self):
        return self.details_open and (
            len(self.pile.contents) == 0
            or self.pile.focus_position > 0
        )

    @details_focused.setter
    def details_focused(self, value):
        if value:
            self.pile.focus_position = len(self.pile.contents)-1
        else:
            self.pile.focus_position = 0

    @property
    def details(self):
        if not getattr(self, "_details", None):

            content = self.table.detail_fn((self.data_source))
            logger.debug(f"open_details: {type(content)}")
            if not content:
                return

            # self.table.header.render( (self.table.width,) )
            indent_width = 0
            visible_count = itertools.count()

            def should_indent(x):
                if (isinstance(self.table.detail_hanging_indent, int)
                    and (x[2] is None or x[2] <= self.table.detail_hanging_indent)):
                    return True
                elif (isinstance(self.table.detail_hanging_indent, str)
                    and x[1].name != self.table.detail_hanging_indent):
                      return True
                return False

            if self.table.detail_hanging_indent:
                indent_width = sum([
                    x[1].width if not x[1].hide else 0
                    for x in itertools.takewhile(
                            should_indent,
                            [ (i, c, next(visible_count) if not c.hide else None)
                              for i, c in enumerate(self.table._columns) ]
                    )
                ])

            self._details = DataTableDetails(self, content, indent_width)
        return self._details


    def open_details(self):

        if not self.table.detail_fn or not self.details or self.details_open:
            return

        if len(self.pile.contents) > 1:
            return

        if self.table.detail_replace:
            self.pile.contents[0] = (urwid.Filler(urwid.Text("")), self.pile.options("given", 0))

        self.pile.contents.append(
            (self.details, self.pile.options("pack"))
        )

        self.details_focused = True
        if not self["_details"]:
            self["_details"] = AttrDict()
        self["_details"]["open"] = True


    def close_details(self):
        if not self.table.detail_fn or not self.details_open:
            return
        # raise Exception
        self["_details"]["open"] = False

        if self.table.detail_replace:
            self.pile.contents[0] = (self.box, self.pile.options("pack"))

            # del self.pile.contents[:]
            # self.pile.contents.append(
            #     (self.box, self.pile.options("pack"))
            # )
        if len(self.pile.contents) >= 2:
            del self.pile.contents[1]

    def toggle_details(self):

        if self.details_open:
            self.close_details()
        else:
            self.open_details()


    def set_attr(self, attr):
        attr_map = self.attrmap.get_attr_map()
        attr_map[self.ATTR] = attr
        if self.cell_selection:
            attr_map[self.attr_highlight] = "%s highlight focused" %(attr)
        else:
            attr_map[self.attr_highlight] = "%s highlight" %(attr)
        self.attrmap.set_attr_map(attr_map)

        focus_map = self.attrmap.get_focus_map()
        focus_map[self.ATTR] = "%s focused" %(attr)
        focus_map[self.attr_highlight] = "%s highlight focused" %(attr)
        if self.cell_selection:
            focus_map[self.attr_focused] = "%s column_focused" %(attr)
            focus_map[self.attr_highlight_focused] = "%s highlight column_focused" %(attr)
        else:
            focus_map[self.attr_focused] = "%s focused" %(attr)
            focus_map[self.attr_highlight_focused] = "%s highlight focused" %(attr)
        self.attrmap.set_focus_map(focus_map)

    def clear_attr(self, attr):
        attr_map = self.attrmap.get_attr_map()
        for a in [self.ATTR, self.attr_highlight]:
            if a in attr_map:
                del attr_map[a]
        self.attrmap.set_attr_map(attr_map)
        focus_map = self.attrmap.get_focus_map()
        for a in [self.attr_focused, self.attr_highlight_focused]:
            if a in focus_map:
                del focus_map[a]
        focus_map[self.ATTR] = "%s focused" %(self.ATTR)
        self.attrmap.set_focus_map(focus_map)

    def make_cells(self):

        def col_to_attr(col):
            if col.attr is None:
                return None
            if callable(col.attr):
                return col.attr(self.data)
            elif col.attr in self.data:
                return self.data[col.attr]
            elif isinstance(col.attr, str):
                return col.attr
            else:
                return None

        return [
            DataTableBodyCell(
                self.table,
                col,
                self,
                # self.data[col.name] if not col.format_record else self.data,
                value_attr=col_to_attr(col),
                cell_selection=self.cell_selection
            )
            if isinstance(col, DataTableColumn)
            else DataTableDividerBodyCell(self.table, col, self)
            for i, col in enumerate(self.table.visible_columns)]

# class DataTableDetailRow(DataTableRow):

#     ATTR = "table_row_detail"

#     @property
#     def details(self):
#         return self.content

#     def make_contents(self):
#         col = DataTableColumn("details")
#         return DataTableDetailCell(self.table, col, self)

#     def selectable(self):
#         return self.table.detail_selectable


class DataTableHeaderRow(DataTableRow):

    signals = ["column_click", "drag"]

    ATTR = "table_row_header"

    DIVIDER_CLASS = DataTableDividerHeaderCell

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mouse_press = False
        self.mouse_dragging = False
        self.mouse_drag_start = None
        self.mouse_drag_end = None
        self.mouse_drag_source = None
        self.mouse_drag_source_column = None

    def make_cells(self):
        cells = [
            DataTableHeaderCell(
                self.table,
                col,
                self,
                sort=self.sort,
            )
            if isinstance(col, DataTableColumn)
            else DataTableDividerHeaderCell(self.table, col, self)
            for i, col in enumerate(self.table.visible_columns)]

        def sort_by_index(source, index):
            urwid.emit_signal(self, "column_click", index)

        if self.table.ui_sort:
            for i, cell in enumerate([c for c in cells if not isinstance(c, DataTableDividerCell)]):
                urwid.connect_signal(
                    cell,
                    'click',
                    functools.partial(sort_by_index, index=i)
                )
                urwid.connect_signal(
                    cell,
                    "select",
                    functools.partial(sort_by_index, index=i)
                )

        return cells

    def selectable(self):
        return self.table.ui_sort

    def update_sort(self, sort):
        for c in self.data_cells:
            c.update_sort(sort)

    def mouse_event(self, size, event, button, col, row, focus):

        if not super().mouse_event(size, event, button, col, row, focus):
            if event == "mouse press":
                self.mouse_press = True
                if self.mouse_drag_start is None:
                    self.mouse_drag_start = col
            elif event == "mouse drag":
                if self.mouse_press:
                    self.mouse_dragging = True
                    # FIXME
                    # self.mouse_drag_end = col
                    # urwid.emit_signal(
                    #     self,
                    #     "drag",
                    #     self.mouse_drag_source,
                    #     self.mouse_drag_source_column,
                    #     self.mouse_drag_start, self.mouse_drag_end
                    # )
                    # self.mouse_dragging = False
                    # self.mouse_drag_start = None
                    # FIXME
            elif event == "mouse release":
                self.mouse_press = False
                if self.mouse_dragging:
                    # raise Exception(f"drag: {self.mouse_drag_source}")
                    self.mouse_dragging = False
                    self.mouse_drag_end = col
                    # raise Exception(self.mouse_drag_start)
                    urwid.emit_signal(
                        self,
                        "drag",
                        self.mouse_drag_source,
                        self.mouse_drag_source_column,
                        self.mouse_drag_start, self.mouse_drag_end
                    )
                    # raise Exception(self.mouse_drag_source.column.name, self.mouse_drag_start, self.mouse_drag_end)
                    self.mouse_drag_source = None
                    self.mouse_drag_start = None



class DataTableFooterRow(DataTableRow):

    ATTR = "table_row_footer"

    DIVIDER_CLASS = DataTableDividerFooterCell

    def make_cells(self):
        return [
            DataTableFooterCell(
                self.table,
                col,
                self,
                sort=self.sort,
            )
            if isinstance(col, DataTableColumn)
            else DataTableDividerBodyCell(self.table, col, self)
            for i, col in enumerate(self.table.visible_columns)]

    def selectable(self):
        return False



import logging
logger = logging.getLogger(__name__)

import urwid
import asyncio

class PopUpMixin(object):

    def open_popup(self, view, title=None, width=75, height=75):

        urwid.connect_signal(
            view, "close_popup", self.close_popup
        )

        popup = PopUpFrame(self, view, title=title)
        overlay = PopUpOverlay(
            self, popup, view,
            'center', ('relative', width),
            'middle', ('relative', height)
        )
        self._w.original_widget = overlay
        self.popup_visible = True

    def close_popup(self, source):
        self._w.original_widget = self.view
        self.popup_visible = False


class PopUpFrame(urwid.WidgetWrap):

    def __init__(self, parent, body, title = None):

        self.parent = parent
        self.line_box = urwid.LineBox(body)
        super(PopUpFrame, self).__init__(self.line_box)


class PopUpOverlay(urwid.Overlay):

    def __init__(self, parent, *args, **kwargs):
        self.parent = parent
        super(PopUpOverlay,self).__init__(*args, **kwargs)

    def keypress(self, size, key):
        key = super().keypress(size, key)
        if key in [ "esc", "q" ]:
            self.parent.close_popup()
        else:
            return key

class BasePopUp(urwid.WidgetWrap):

    signals = ["close_popup"]

    def selectable(self):
        return True

class ChoiceDialog(BasePopUp):

    choices = []
    signals = ["select"]

    def __init__(self, parent, prompt=None):
        self.parent = parent
        if prompt: self.prompt = prompt
        self.text = urwid.Text(
            self.prompt + " [%s]" %("".join(list(self.choices.keys()))), align="center"
        )
        super(ChoiceDialog, self).__init__(
            urwid.Filler(urwid.Padding(self.text))
        )

    @property
    def choices(self):
        raise NotImplementedError

    def keypress(self, size, key):
        if key in list(self.choices.keys()):
            self.choices[key]()
            self._emit("select", key)
        else:
            return key


class SquareButton(urwid.Button):

    button_left = urwid.Text("[")
    button_right = urwid.Text("]")

    def pack(self, size, focus=False):
        cols = sum(
            [ w.pack()[0] for w in [
                self.button_left,
                self._label,
                self.button_right
            ]]) + self._w.dividechars*2

        return ( cols, )

class OKCancelDialog(BasePopUp):

    def __init__(self, parent, *args, **kwargs):

        self.parent = parent

        self.ok_button = SquareButton(("bold", "OK"))

        urwid.connect_signal(
            self.ok_button, "click",
            lambda s: self.confirm()
        )

        self.cancel_button = SquareButton(("bold", "Cancel"))

        urwid.connect_signal(
            self.cancel_button, "click",
            lambda s: self.cancel()
        )


        self.body = urwid.Pile([])
        for name, widget in self.widgets.items():
            setattr(self, name, widget)
            self.body.contents.append(
                (widget, self.body.options("weight", 1))
            )

        self.pile = urwid.Pile(
            [
                (2, urwid.Filler(urwid.Padding(self.body), valign="top")),
                ("weight", 1, urwid.Padding(
                    urwid.Columns([
                        ("weight", 1,
                         urwid.Padding(
                             self.ok_button, align="center", width=12)
                         ),
                        ("weight", 1,
                         urwid.Padding(
                             self.cancel_button, align="center", width=12)
                         )
                    ]),
                    align="center"
                )),
            ]
        )
        self.body_position = 0
        if self.title:
            self.pile.contents.insert(
                0,
                (urwid.Filler(
                    urwid.AttrMap(
                        urwid.Padding(
                            urwid.Text(self.title)
                        ),
                        "header"
                    )
                ), self.pile.options("given", 2))
            )
            self.body_position += 1

        self.pile.selectable = lambda: True
        self.pile.focus_position = self.body_position
        super(OKCancelDialog, self).__init__(
            urwid.Filler(self.pile, valign="top")
        )

    @property
    def title(self):
        return None

    @property
    def widgets(self):
        raise RuntimeError("must set widgets property")

    def action(self):
        raise RuntimeError("must override action method")

    @property
    def ok_focus_path(self):
        return [self.body_position+1,0]

    @property
    def cancel_focus_path(self):
        return [self.body_position+1,1]

    @property
    def focus_paths(self):
        return [
            [self.body_position, i]
            for i in range(len(self.body.contents))
        ] + [
            self.ok_focus_path,
            self.cancel_focus_path
        ]

    def cycle_focus(self, step):
        path = self.pile.get_focus_path()[:2]
        logger.info(f"{path}, {self.focus_paths}")
        self.pile.set_focus_path(
            self.focus_paths[
                (self.focus_paths.index(path) + step) % len(self.focus_paths)
            ]
        )

    def confirm(self):
        rv = self.action()
        if asyncio.iscoroutine(rv):
            asyncio.get_event_loop().create_task(rv)

        self.close()

    def cancel(self):
        self.close()

    def close(self):
        self._emit("close_popup")

    def selectable(self):
        return True

    def keypress(self, size, key):
        if key in ["tab", "shift tab"]:
            self.cycle_focus(1 if key == "tab" else -1)
            return
        else:
            key = super().keypress(size, key)
            if key == "enter":
                self.confirm()
            else:
                return key



class ConfirmDialog(ChoiceDialog):

    def __init__(self, parent, *args, **kwargs):
        super(ConfirmDialog, self).__init__(parent, *args, **kwargs)

    def action(self, value):
        raise RuntimeError("must override action method")

    @property
    def prompt(self):
        return "Are you sure?"

    def confirm(self):
        self.action()
        self.close()

    def cancel(self):
        self.close()

    def close(self):
        self.parent.close_popup()

    @property
    def choices(self):
        return {
            "y": self.confirm,
            "n": self.cancel
        }

class BaseView(urwid.WidgetWrap):

    focus_widgets = []
    top_view = None

    def __init__(self, view):

        self.view = view
        self.placeholder = urwid.WidgetPlaceholder(urwid.Filler(urwid.Text("")))
        super(BaseView, self).__init__(self.placeholder)
        self.placeholder.original_widget = self.view

    def open_popup(self, view, title=None, width=("relative", 75), height=("relative", 75)):

        urwid.connect_signal(
            view, "close_popup", self.close_popup
        )

        popup = PopUpFrame(self, view, title=title)
        overlay = PopUpOverlay(
            self, popup, self.view,
            'center', width,
            'middle', height
        )
        self._w.original_widget = overlay
        self.popup_visible = True

    def close_popup(self, source=None):
        self._w.original_widget = self.view
        self.popup_visible = False

__all__ = [
    "BaseView",
    "BasePopUp",
    "ChoiceDialog",
]



import logging
logger = logging.getLogger("panwid.datatable")

from .common import *

import urwid
class DataTableCell(urwid.WidgetWrap):

    signals = ["click", "select"]

    ATTR = "table_cell"

    def __init__(self, table, column, row,
                 fill=False,
                 value_attr=None,
                 cell_selection=False,
                 padding=0,
                 *args, **kwargs):


        self.table = table
        self.column = column
        self.row = row

        self.fill = fill
        self.value_attr = value_attr
        self.cell_selection = cell_selection

        self.attr = self.ATTR
        self.attr_focused = "%s focused" %(self.attr)
        self.attr_column_focused = "%s column_focused" %(self.attr)
        self.attr_highlight = "%s highlight" %(self.attr)
        self.attr_highlight_focused = "%s focused" %(self.attr_highlight)
        self.attr_highlight_column_focused = "%s column_focused" %(self.attr_highlight)

        self._width = None
        self._height = None
        self.contents_rows = None
        # self.width = None

        if column.padding:
            self.padding = column.padding
        else:
            self.padding = padding

        self.update_contents()

        # logger.info(f"{self.column.name}, {self.column.width}, {self.column.align}")
        # if self.row.row_height is not None:

        # self.filler = urwid.Filler(self.contents)

        self.normal_attr_map = {}
        self.highlight_attr_map = {}

        self.normal_focus_map = {}
        self.highlight_focus_map = {}
        self.highlight_column_focus_map = {}

        self.set_attr_maps()

        self.highlight_attr_map.update(self.table.highlight_map)
        self.highlight_focus_map.update(self.table.highlight_focus_map)

        self.attrmap = urwid.AttrMap(
            # self.filler,
            urwid.Filler(self.contents) if "flow" in self.contents.sizing() else self.contents,
            attr_map = self.normal_attr_map,
            focus_map = self.normal_focus_map
        )
        super(DataTableCell, self).__init__(self.attrmap)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.column.name}>"

    @property
    def value(self):
        if self.column.value_fn:
            val = self.column.value_fn(self.table, self.row)
        else:
            val = self.row[self.column.name]
        return val

    @value.setter
    def value(self, value):
        self.table.df[self.row.index, self.column.name] = value

    @property
    def formatted_value(self):

        v = self.column._format(self.value)
        if not self.width:
            return v
        # try:
        v = str(v)[:self.width-self.padding*2]
        # logger.info(f"formatted_value: {v}")
        return v
        # except TypeError:
            # raise Exception(f"{v}, {type(v)}")

    def update_contents(self):
        pass

    def set_attr_maps(self):

        self.normal_attr_map[None] = self.attr
        self.highlight_attr_map [None] = self.attr_highlight
        self.normal_focus_map[None] = self.attr_focused
        self.highlight_focus_map[None] = self.attr_highlight_focused

        if self.value_attr:
            self.normal_attr_map.update({None: self.value_attr})
            self.normal_focus_map.update({None: "%s focused" %(self.value_attr)})
            self.highlight_attr_map.update({None: "%s highlight" %(self.value_attr)})
            if self.cell_selection:
                self.highlight_focus_map.update({None: "%s highlight column_focused" %(self.value_attr)})
            else:
                self.highlight_focus_map.update({None: "%s highlight focused" %(self.value_attr)})

    def highlight(self):
        self.attrmap.set_attr_map(self.highlight_attr_map)
        self.attrmap.set_focus_map(self.highlight_focus_map)

    def unhighlight(self):
        self.attrmap.set_attr_map(self.normal_attr_map)
        self.attrmap.set_focus_map(self.normal_focus_map)

    def enable_selection(self):
        self.cell_selection = True

    def disable_selection(self):
        self.cell_selection = False

    def selectable(self):
        return self.cell_selection

    def keypress(self, size, key):
        try:
            key = super(DataTableCell, self).keypress(size, key)
        except AttributeError:
            pass
        return key
        # return super(DataTableCell, self).keypress(size, key)

    def set_attr_map(self, attr_map):
        self.attrmap.set_attr_map(attr_map)

    def mouse_event(self, size, event, button, col, row, focus):
        if event == 'mouse press':
            urwid.emit_signal(self, "click")

    def set_attr(self, attr):
        attr_map = self.attrmap.get_attr_map()
        attr_map[None] = attr
        # self.attrmap.set_attr_map(attr_map)
        focus_map = self.attrmap.get_focus_map()
        focus_map[None] = "%s focused" %(attr)
        self.attrmap.set_focus_map(focus_map)

    def clear_attr(self, attr):
        attr_map = self.attrmap.get_attr_map()
        attr_map = self.normal_attr_map
        # attr_map[None] = self.attr
        self.attrmap.set_attr_map(attr_map)
        focus_map = self.normal_focus_map #.attr.get_focus_map()
        # focus_map[None] = self.attr_focused
        self.attrmap.set_focus_map(focus_map)

    def render(self, size, focus=False):
        # logger.info("cell render")
        maxcol = size[0]
        self._width = size[0]
        if len(size) > 1:
            maxrow = size[1]
            self._height = maxrow
        else:
            self.contents_rows = self.contents.rows(size, focus)
            self._height = contents_rows

        if getattr(self.column, "truncate", None):
            rows = self.inner_contents.pack((self.width,))[1]
            if rows > 1:
                self.truncate()
        return super().render(size, focus)

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def inner_contents(self):
        return self.contents

    def truncate(self):
        pass


class DataTableDividerCell(DataTableCell):

    # @property
    # def fiil(self):
    #     return self.row.row_height is not None

    def selectable(self):
        return False

    def update_contents(self):
        if not (
                (isinstance(self, DataTableHeaderCell) and not self.column.in_header)
                or (isinstance(self, DataTableFooterCell) and not self.column.in_footer)
        ):
            divider = self.column.value
        else:
            divider = urwid.Divider(" ")
        contents = urwid.Padding(
            divider,
            left = self.column.padding_left,
            right = self.column.padding_right
        )
        self.contents = contents
        # self._invalidate()

class DataTableBodyCell(DataTableCell):

    ATTR = "table_row_body"

    def update_contents(self):

        self.inner = self.table.decorate(
            self.row,
            self.column,
            self.formatted_value
        )

        # try:
        #     self.inner = self.table.decorate(
        #         self.row,
        #         self.column,
        #         self.formatted_value
        #     )
        # except Exception as e:
        #     logger.exception(e)
        #     self.inner = urwid.Text("")

        if getattr(self.column, "truncate", None):
            end_char = u"\N{HORIZONTAL ELLIPSIS}" if self.column.truncate is True else self.column.truncate
            contents = urwid.Columns([
                ("weight", 1, self.inner),
                (0, urwid.Text(end_char))
            ])
            contents = urwid.Filler(contents)
        else:
            width = "pack"

            contents = urwid.Padding(
                self.inner,
                align=self.column.align,
                width = width,
                left=self.padding,
                right=self.padding
            )
            # contents = urwid.Filler(contents)

        self.contents = contents

    def truncate(self):
        columns = self.contents.original_widget
        col = columns.contents[1]
        col = (col[0], columns.options("given", 1))
        del self.contents.original_widget.contents[1]
        columns.contents.append(col)


    @property
    def inner_contents(self):
        return self.inner


class DataTableDividerBodyCell(DataTableDividerCell, DataTableBodyCell):
    pass

class DataTableHeaderCell(DataTableCell):

    ATTR = "table_row_header"

    ASCENDING_SORT_MARKER = u"\N{UPWARDS ARROW}"
    DESCENDING_SORT_MARKER = u"\N{DOWNWARDS ARROW}"

    def __init__(self, *args, **kwargs):
        self.mouse_dragging = False
        self.mouse_drag_start = None
        self.mouse_drag_end = None
        super().__init__(*args, **kwargs)

    @property
    def index(self):
        return next(i for i, c in enumerate(self.table.visible_columns)
                    if c.name == self.column.name)

    @property
    def min_width(self):
        return len(self.label) + self.padding*2 + (1 if self.sort_icon else 0)

    def update_contents(self):

        self.label = self.column.label
        if self.column.sort_icon is not None:
            self.sort_icon = self.column.sort_icon
        else:
            self.sort_icon = self.table.sort_icons

        label = (self.label
                 if isinstance(self.label, urwid.Widget)
                 else
                 DataTableText(
                     self.label,
                     wrap = "space" if self.column.no_clip_header else "clip",
                 )
        )

        padding = urwid.Padding(
            label,
            align=self.column.align,
            width="pack",
            left=self.padding,
            right=self.padding
        )

        columns = urwid.Columns([
            ("weight", 1, padding)
        ])

        if self.sort_icon:
            if self.column.align == "right":
                columns.contents.insert(0,
                    (DataTableText(""), columns.options("given", 1))
                )
            else:
                columns.contents.append(
                    (DataTableText(""), columns.options("given", 1))
                )
        self.contents = columns

        # self.contents = DataTableText(
        #     self.label,
        #     wrap = "space" if self.column.no_clip_header else "clip"
        # )
        self.update_sort(self.table.sort_by)

    def set_attr_maps(self):

        self.normal_attr_map[None] = self.attr
        self.highlight_attr_map [None] = self.attr_highlight
        # if self.cell_selection:
        self.normal_focus_map[None] = self.attr_column_focused
        self.highlight_focus_map[None] = self.attr_highlight_column_focused

    def selectable(self):
        return self.table.ui_sort

    def keypress(self, size, key):
        if key != "enter":
            return key
        urwid.emit_signal(self, "select", self)

    def mouse_event(self, size, event, button, col, row, focus):
        if event == "mouse press":
            logger.info("cell press")
            if self.mouse_drag_start is None:
                self.row.mouse_drag_source_column = col
            self.row.mouse_drag_source = self
            return False
        elif event == "mouse drag":
            logger.info("cell drag")
            self.mouse_dragging = True
            return False
                # urwid.emit_signal(self, "drag_start")
        elif event == "mouse release":
            logger.info("cell release")
            if self.mouse_dragging:
                self.mouse_dragging = False
                self.mouse_drag_start = None
            else:
                urwid.emit_signal(self, "select", self)
            self.mouse_drag_source = None
        #     self.mouse_drag_end = col
        #     raise Exception(self.mouse_drag_start, self.mouse_drag_end)
        #     self.mouse_drag_start = None
        super().mouse_event(size, event, button, col, row, focus)

    def update_sort(self, sort):
        if not self.sort_icon: return

        index = 0 if self.column.align=="right" else 1
        if sort and sort[0] == self.column.name:
            direction = self.DESCENDING_SORT_MARKER if sort[1] else self.ASCENDING_SORT_MARKER
            self.contents.contents[index][0].set_text(direction)
        # else:
        #     self.contents.contents[index][0].set_text("")


class DataTableDividerHeaderCell(DataTableDividerCell, DataTableHeaderCell):
    pass


class DataTableFooterCell(DataTableCell):

    ATTR = "table_row_footer"

    def update_contents(self):
        if self.column.footer_fn and len(self.table.df):
            # self.table.df.log_dump()
            if self.column.footer_arg == "values":
                footer_arg = self.table.df[self.column.name].to_list()
            elif self.column.footer_arg == "rows":
                footer_arg = self.table.df.iterrows()
            elif self.column.footer_arg == "table":
                footer_arg = self.table.df
            else:
                raise Exception
            self.contents = self.table.decorate(
                self.row,
                self.column,
                self.column._format(self.column.footer_fn(self.column, footer_arg))
            )
        else:
            self.contents = DataTableText("")


class DataTableDividerFooterCell(DataTableDividerCell, DataTableHeaderCell):

    DIVIDER_ATTR = "table_divider_footer"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import sys
from os import path
from glob import glob

name = 'panwid'
setup(name=name,
      version='0.3.4',
      description='Useful widgets for urwid',
      author='Tony Cebzanov',
      author_email='tonycpsu@gmail.com',
      url='https://github.com/tonycpsu/panwid',
      python_requires='>=3.6',
      classifiers=[
          'Environment :: Console',
          'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
          'Intended Audience :: Developers'
      ],
      packages=find_packages(),
      data_files=[('share/doc/%s' % name, ['LICENSE','README.md']),
              ],
      install_requires = [
          "urwid",
          "urwid-utils>=0.1.2",
          "six",
          "raccoon>=3.0.0",
          "orderedattrdict"
      ],
      test_suite="test",
      # dependency_links=[
      #     "https://github.com/tonycpsu/urwid_utils/tarball/master#egg=urwid_utils-0.0.5dev"
      # ],
     )
